# -*- coding: utf-8 -*-

from datajuggler.serializer.abstract import AbstractSerializer

try:
    try:
        import tomllib as toml   # python 3.11 or later.
    except ImportError:
        import toml


    class TOMLSerializer(AbstractSerializer):
        """
        This class describes a toml serializer.
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            return toml.loads(s, **kwargs)

        def encode(self, d, **kwargs):
            return toml.dumps(dict(d), **kwargs)


except ImportError:

    class TOMLSerializer(AbstractSerializer):
        """
        This class describes a toml serializer.
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            raise NotImplementedError('You should install toml.')

        def encode(self, d, **kwargs):
            raise NotImplementedError('You should install toml.')



# -*- coding: utf-8 -*-

from datajuggler.serializer.abstract import AbstractSerializer
from datajuggler.serializer.json import JSONSerializer
from pathlib import Path
from typing import Optional, Type, Union

try:
    import yaml
    from yaml.representer import Representer, SafeRepresenter

    def yaml_initializer(cls, factory: Optional[Type[dict]]=None):
        """class method to initialize for YAML"""

        def _from_yaml(loader, node):
            nonlocal factory
            value = loader.construct_mapping(node)
            data = factory(value)
            yield data

        def _to_yaml_safe(dumper, data):
            return dumper.represent_dict(data)

        def _to_yaml(dumper, data):
            nonlocal module_name, factory_name
            return dumper.represent_mapping(u'!{}.{}'.format(module_name, factory_name), data)

        factory = factory or cls
        factory_name=factory.__name__
        module_name = Path(__file__).parent.parent.name

        for loader_name in ( "BaseLoader", "FullLoader", "SafeLoader",
                             "Loader", "UnsafeLoader", "DangerLoader" ):
            LoaderCls = getattr(yaml, loader_name, None)
            if LoaderCls is None:
                # This code supports both PyYAML 4.x and 5.x versions
                continue
            yaml.add_constructor(u'!{}'.format( module_name ),
                                  _from_yaml, Loader=LoaderCls)
            yaml.add_constructor(u'!{}.{}'.format( module_name, factory_name ),
                                  _from_yaml, Loader=LoaderCls)
            yaml.add_constructor(u'!python/object:{}.{}'.format(
                                           module_name, factory_name),
                                  _from_yaml, Loader=LoaderCls)
            yaml.add_constructor(u'!python/object/new:{}.{}'.format(
                                           module_name, factory_name),
                                  _from_yaml, Loader=LoaderCls)

        SafeRepresenter.add_representer(cls, _to_yaml_safe)
        SafeRepresenter.add_multi_representer(cls, _to_yaml_safe)

        Representer.add_representer(cls, _to_yaml)
        Representer.add_multi_representer(cls, _to_yaml)


    class YAMLSerializer(AbstractSerializer):
        """
        This class describes an yaml serializer.
        """

        def __init__(self, factory: Optional[Type[dict]]=None):
            super().__init__()
            self._json_serializer = JSONSerializer()

        def decode(self, s, **kwargs):
            factory = lambda d: dict(*(args + (d,)), **kwargs)
            loader_class = kwargs.pop('Loader', yaml.FullLoader)
            return yaml.load(s, Loader=loader_class)

        def encode(self, d, **options):
            d = self._json_serializer.decode(self._json_serializer.encode(d))
            opts = dict(indent=4, default_flow_style=False, allow_unicode=True)
            opts.update(options)
            if 'Dumper' not in opts:
                return yaml.safe_dump(d, **opts)
            else:
                return yaml.dump(d, **opts)
            return data

except ImportError:
    def yaml_initializer(cls, factory: Optional[Type[dict]]=None):
        pass

    class YAMLSerializer(AbstractSerializer):
        """
        This class describes an yaml serializer.
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            raise NotImplementedError('You should install PyYAML.')

        def encode(self, d, **kwargs):
            raise NotImplementedError('You should install PyYAML.')




# -*- coding: utf-8 -*-

from datajuggler.serializer.abstract import AbstractSerializer

try:
    import xmltodict

    class XMLSerializer(AbstractSerializer):
        """
        This class describes a xml serializer.
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            kwargs.setdefault("dict_constructor", dict)
            data = xmltodict.parse(s, **kwargs)
            return data

        def encode(self, d, **kwargs):
            if len(list(d.keys())) != 1:
                raise ValueError('dict must have exactly one root.')
            data = xmltodict.unparse(d, **kwargs)
            return data

except ImportError:

    class XMLSerializer(AbstractSerializer):
        """
        This class describes a xml serializer.
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            raise NotImplementedError('You should install xmltodict.')

        def encode(self, d, **kwargs):
            raise NotImplementedError('You should install xmltodict.')



# -*- coding: utf-8 -*-

from __future__ import annotations

import re
from typing import (
    Any, Dict, Union, Optional, Hashable, Iterable, Sequence,
    Callable, Pattern, Type
    )
import copy
import json
from datajuggler.strings import is_match_string, searchstr

from datajuggler.validator import (
    DictAction, DictActionType,
    DictItem, DictItemType,
    DictKey, DictKeyType,
    )
from datajuggler.keys import (
    Keylist, Keypath, Default_Keypath_Separator
    )

from datajuggler import dicthelper as d
from datajuggler import serializer as io
from datajuggler.validator import TypeValidator as _type

class BaseDict(dict):

    """ Factory class for custom dictionary """
    def __init__(self,
            *args: Any,
            **kwargs: Any
        ):
        super().__init__(*args, **kwargs)


    def __repr__(self):
        return f'{self.__class__.__name__}({dict.__repr__(self)})'

    def __str__(self):
        return f'{dict.__repr__(self)}'

    def __getstate__(self):
        return {k: v for k, v in self.items()}

    def __setstate__(self, state):
        self.clear()
        self.update(state)

    def __deepcopy__(self, memo):
        other = self.__class__()
        memo[id(self)] = other
        for key, value in self.items():
            other[copy.deepcopy(key, memo)] = copy.deepcopy(value, memo)
        return other

    def copy(self,
            factory: Optional[Type[dict]]=None,
        ):
        if factory:
            return factory(copy.copy(self))
        else:
            return copy.copy(self)

    def deepcopy(self,
            factory: Optional[Type[dict]]=None,
        ):
        if factory:
            return factory(copy.deepcopy(self))
        else:
            return copy.deepcopy(self)

    def update(self, *args, **kwargs):
        for key, val in dict(*args, **kwargs).items():
            self[key] = val

    def get(self, key: Hashable, default=None):
        if key not in self:
            return default
        return self[key]

    def setdefault(self, key: Hashable, default=None):
        if key not in self:
            self[key] = default
        return self[key]

    def fromkeys(self,
            seq: Sequence,
            value: Any,
            inplace: bool=False,
        ) ->Optional[dict]:
        """Create a new dictionary with keys from iterable and values set to value.
           If set `True` to `inplace`, perform operation in-place.
        """
        new = type(self)(dict(self).fromkeys(seq, value))
        if inplace:
            self.clear()
            self.update(new)
        else:
            return new

    def fromvalues(self,
            seq: Sequence,
            base: int=1,
            prefix: Optional[str]=None,
            inplace: bool=False,
        ) ->Optional[dict]:
        """Create a new dictionary from list of values.
           keys automaticaly generate as interger.
           `base` is the starting number.
           If set `True` to `inplace`, perform operation in-place.
        """
        if prefix != None:
            new = type(self)({'{}{}'.format(prefix, base+x): seq[x]
                                            for x in range(len(seq)) } )
        else:
            new = type(self)({base+x: seq[x] for x in range(len(seq))})
        if inplace:
            self.update(new)
        else:
            return new

    def fromlists(self,
            keys: Sequence,
            values: Sequence,
            inplace: bool=False,
        ) ->Optional[dict]:
        """Create a new dictionary from two list as keys and values.
           Only the number of elements in the shorter of the two lists is processed.
           If set `True` to `inplace`, perform operation in-place.
        """
        zipobj = zip(keys, values)
        new = type(self)(dict(zipobj))
        if inplace:
            self.update(new)
        else:
            return new

    def to_dict(self, obj: Optional[dict]=None):
        """ Recursively converts BaseDict to dict.  """
        obj = obj if obj or obj == {} else self
        holding_obj = dict()

        def convert_loop(obj):
            try:
                return holding_obj[id(obj)]
            except KeyError:
                pass

            holding_obj[id(obj)] = partial = pre_convert(obj)
            return post_convert(partial, obj)

        def pre_convert(obj):
            if _type.is_dict(obj):
                return dict()
            elif _type.is_list(obj):
                return type(obj)()
            elif _type.is_tuple(obj):
                type_factory = getattr(obj, "_make", type(obj))
                return type_factory(convert_loop(item) for item in obj)
            else:
                return obj

        def post_convert(partial, obj):
            if _type.is_dict(obj):
                partial.update((k, convert_loop(obj[k])) for k in obj.keys())
            elif _type.is_list(obj):
                partial.extend(convert_loop(v) for v in obj)
            elif _type.is_tuple(obj):
                for (value_partial, value) in zip(partial, obj):
                    post_convert(value_partial, value)

            return partial

        return convert_loop(obj)

    def from_dict(self,
            obj: Any,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ):
        """ Recursively converts from dict to BaseDict. """
        factory = factory or type(self)
        holding_obj = dict()
        workdict = type(self)()

        def convert_loop(obj):
            try:
                return holding_obj[id(obj)]
            except KeyError:
                pass

            holding_obj[id(obj)] = partial = pre_convert(obj)
            return post_convert(partial, obj)

        def pre_convert(obj):
            if _type.is_dict(obj):
                return factory({})
            elif _type.is_list(obj):
                return type(obj)()
            elif _type.is_tuple(obj):
                type_factory = getattr(obj, "_make", type(obj))
                return type_factory(convert_loop(item) for item in obj)
            else:
                return obj

        def post_convert(partial, obj):
            if _type.is_dict(obj) and factory == type(self):
                partial = factory((key, convert_loop(obj[key]))
                                for key in obj.keys() )
            elif _type.is_dict(obj):
                partial.update((key, convert_loop(obj[key]))
                                for key in obj.keys() )
            elif _type.is_list(obj):
                partial.extend(convert_loop(item) for item in obj)
            elif _type.is_tuple(obj):
                for (item_partial, item) in zip(partial, obj):
                    post_convert(item_partial, item)

            return partial

        obj = convert_loop(obj)
        try:
            if inplace:
                self.update(obj)
            else:
                workdict.update(obj)

        except AttributeError:
            pass   # obj is not Mapping

        if not inplace:
            return workdict


class IODict(BaseDict):
    yaml_initializer = classmethod(io.yaml_initializer)

    def __init__(self,
            *args: Any,
            format: Optional[str]=None,
            **kwargs: Any):
        """
        Constructs a new instance.
        """
        # if first argument is data-string, url or filepath try to decode it.
        # use 'format' kwarg to specify the decoder to use, default 'json'.
        if format:
            if  format.lower() in ["yml", "yaml"]:
                self.yaml_initializer()
            d = IODict._decode_init(*args, format=format, **kwargs)
            super().__init__(d, **kwargs)
            return
        super().__init__(*args, **kwargs)

    @staticmethod
    def _serializer_init(s, **kwargs):
        pass

    @staticmethod
    def _decode_init(s,
            format: Optional[str]=None,
            **kwargs: Any
        ) ->dict:
        autodetected_format = io.autodetect_format(s)
        default_format = autodetected_format or "json"
        # format = kwargs.pop("format", default_format).lower()
        format = format or default_format
        if format in ["b64", "base64"]:
            kwargs.setdefault("subformat", "json")
        # decode data-string and initialize with dict data.
        return IODict._decode(s, format=format, **kwargs)

    @staticmethod
    def _decode(s,
            format: str,
            **kwargs: Any
        ) ->dict:
        try:
            content = io.read_contents(s)
            # decode content using the given format
            data = io.decode(content, format=format, **kwargs)
            if _type.is_dict(data):
                return data
            elif _type.is_list(data):
                # force list to dict
                return {"values": data}
            else:
                raise ValueError(
                    f"Invalid data type: {type(data)}, expected dict or list."
                )
        except Exception as e:
            raise ValueError(f"Invalid data or url or filepath argument: {s}\n{e}")

    @staticmethod
    def _encode(d, format, **kwargs):
        filepath = kwargs.pop("filepath", None)
        s = io.encode(d, format, **kwargs)
        if filepath:
            io.write_file(filepath, s)
        return s

    @classmethod
    def from_base64(cls, s, subformat="json", encoding="utf-8", **kwargs):
        """
        Load and decode Base64 data from url, filepath or data-string.
        Data is decoded according to subformat and encoding.
        Decoder specific options can be passed using kwargs.
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        kwargs["subformat"] = subformat
        kwargs["encoding"] = encoding
        return cls(s, format="base64", **kwargs)

    @classmethod
    def from_csv(cls, s, columns=None, columns_row=True, **kwargs):
        """
        Load and decode CSV data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/csv.html
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        kwargs["columns"] = columns
        kwargs["columns_row"] = columns_row
        return cls(s, format="csv", **kwargs)

    @classmethod
    def from_ini(cls, s, **kwargs):
        """
        Load and decode INI data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/configparser.html
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="ini", **kwargs)

    def from_json(self,
            stream,
            inplace:bool=False,
            **options: Any):
        """
        Load and decode JSON data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/json.html
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        if inplace:
            self.update(json.loads(stream, **options))
        else:
            return type(self)(json.loads(stream, **options))

    @classmethod
    def from_pickle(cls, s, **kwargs):
        """
        Load and decode a pickle encoded in Base64 format data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/pickle.html
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="pickle", **kwargs)

    @classmethod
    def from_plist(cls, s, **kwargs):
        """
        Load and decode p-list data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/plistlib.html
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="plist", **kwargs)

    @classmethod
    def from_query_string(cls, s, **kwargs):
        """
        Load and decode query-string from url, filepath or data-string.
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="query_string", **kwargs)

    @classmethod
    def from_toml(cls, s, **kwargs):
        """
        Load and decode TOML data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://pypi.org/project/toml/
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="toml", **kwargs)

    @classmethod
    def from_xml(cls, s, **kwargs):
        """
        Load and decode XML data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://github.com/martinblech/xmltodict
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="xml", **kwargs)

    @classmethod
    def from_yaml(cls, s, **kwargs):
        """
        Load and decode YAML data from url, filepath or data-string.
        Decoder specific options can be passed using kwargs:
        https://pyyaml.org/wiki/PyYAMLDocumentation
        Return a new dict instance. A ValueError is raised in case of failure.
        """
        return cls(s, format="yaml", **kwargs)

    def to_base64(self, subformat="json", encoding="utf-8", **kwargs):
        """
        Encode the current dict instance in Base64 format
        using the given subformat and encoding.
        Encoder specific options can be passed using kwargs.
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        kwargs["subformat"] = subformat
        kwargs["encoding"] = encoding
        return self._encode(self.to_dict(), "base64", **kwargs)

    def to_csv(self, key=None, columns=None, columns_row=True, **kwargs):
        """
        Encode a list of dicts in the current dict instance in CSV format.
        Encoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/csv.html
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        kwargs["columns"] = columns
        kwargs["columns_row"] = columns_row
        d = self.to_dict()
        key = key or list(d.keys())[0]
        return self._encode(d[key], "csv", **kwargs)

    def to_ini(self, **kwargs):
        """
        Encode the current dict instance in INI format.
        Encoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/configparser.html
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "ini", **kwargs)

    def to_json(self, **kwargs):
        """
        Encode the current dict instance in JSON format.
        Encoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/json.html
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "json", **kwargs)

    def to_pickle(self, **kwargs):
        """
        Encode the current dict instance as pickle (encoded in Base64).
        The pickle protocol used by default is 2.
        Encoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/pickle.html
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "pickle", **kwargs)

    def to_plist(self, **kwargs):
        """
        Encode the current dict instance as p-list.
        Encoder specific options can be passed using kwargs:
        https://docs.python.org/3/library/plistlib.html
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "plist", **kwargs)

    def to_query_string(self, **kwargs):
        """
        Encode the current dict instance in query-string format.
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "query_string", **kwargs)

    def to_toml(self, **kwargs):
        """
        Encode the current dict instance in TOML format.
        Encoder specific options can be passed using kwargs:
        https://pypi.org/project/toml/
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "toml", **kwargs)

    def to_xml(self, **kwargs):
        """
        Encode the current dict instance in XML format.
        Encoder specific options can be passed using kwargs:
        https://github.com/martinblech/xmltodict
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "xml", **kwargs)

    def to_yaml(self, **kwargs):
        """
        Encode the current dict instance in YAML format.
        Encoder specific options can be passed using kwargs:
        https://pyyaml.org/wiki/PyYAMLDocumentation
        Return the encoded string and optionally save it at 'filepath'.
        A ValueError is raised in case of failure.
        """
        return self._encode(self.to_dict(), "yaml", **kwargs)


class aDict(IODict):

    def __init__(__self,
            *args: Any,
            format: Optional[str]=None,
            **kwargs: Any
        ):
        object.__setattr__(__self, '__parent', kwargs.pop('__parent', None))
        object.__setattr__(__self, '__key', kwargs.pop('__key', None))
        object.__setattr__(__self, '__frozen', False)
        if format:
            if format.lower() in ["yml", "yaml"]:
                io.yaml_initializer(cls=aDict)
            args = (super()._decode_init(*args, format=format, **kwargs),)
        else:
            if args and isinstance(args[0], str) and io.validate_file(args[0]):
                format = io.autodetect_format(args[0])
                args = (super()._decode_init(*args, format=format, **kwargs),)

        for arg in args:
            if not arg:
                continue
            elif _type.is_dict(arg):
                for key, val in arg.items():
                    __self[key] = __self._hook(val)
            elif _type.is_tuple(arg)  and (not _type.is_tuple(arg[0])):
                __self[arg[0]] = __self._hook(arg[1])
            else:
                for key, val in iter(arg):
                    __self[key] = __self._hook(val)

        for key, val in kwargs.items():
            __self[key] = __self._hook(val)

    def _check_frozen(self,
            thrown_error: bool=False,
            msg: str='frozen object cannot be modified.',
            ):
        if object.__getattribute__(self, '__frozen'):
            if thrown_error:
                raise AttributeError( f"{self.__class__.__name__} {msg}" )
            else:
                return True
        else:
            return False

    def __setattr__(self, name, value):
        if hasattr(self.__class__, name):
            raise AttributeError("'aDict' object attribute "
                                 "'{0}' is read-only".format(name))
        else:
            self[name] = value

    def __setitem__(self, name, value):
        if self._check_frozen(thrown_error=False):
            if name not in super().keys():
                raise KeyError(name)
            else:
                raise AttributeError("'aDict' object attribute "
                                 "'{0}' is read-only".format(name))

        super().__setitem__(name, value)
        try:
            p = object.__getattribute__(self, '__parent')
            key = object.__getattribute__(self, '__key')
        except AttributeError:
            p = None
            key = None
        if p is not None:
            p[key] = self
            object.__delattr__(self, '__parent')
            object.__delattr__(self, '__key')


    def __add__(self, other):
        if not self.keys():
            return other
        else:
            self_type = type(self).__name__
            other_type = type(other).__name__
            msg = "unsupported operand type(s) for +: '{}' and '{}'"
            raise TypeError(msg.format(self_type, other_type))

    @classmethod
    def _hook(cls, item):
        if _type.is_dict(item):
            return cls(item)
        elif _type.is_list_or_tuple(item):
            return type(item)(cls._hook(elem) for elem in item)
        return item

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __missing__(self, name):
        if object.__getattribute__(self, '__frozen'):
            raise KeyError(name)
        return self.__class__(__parent=self, __key=name)

    def __delattr__(self, name):
        self._check_frozen(thrown_error=True)
        del self[name]

    def __hash__(self):
        if self._check_frozen(thrown_error=False):
            return hash(frozenset(self.items()))
        else:
            raise AttributeError('unhashable not frozen object.')


    def __getnewargs__(self):
        return tuple(self.items())

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self._check_frozen(thrown_error=True)
        self.update(state)

    def __or__(self, other):
        return d.d_merge(self, other,
                         overwrite=True, concat=True,
                         inplace=False, factory=type(self))

    def __ror__(self, other):
        return d.d_merge(other, self,
                         overwrite=True, concat=True,
                         inplace=False, factory=type(self))

    def __ior__(self, other):
        self._check_frozen(thrown_error=True)
        self.update(other)
        return self

    def __str__(self):
        return f'{self.to_dict(self)}'

    def __getattr__(self, k):
        try:
            return object.__getattribute__(self, k)
        except AttributeError:
            try:
                return self[k]
            except KeyError:
                raise AttributeError(k)

    def __setattr__(self, k, v):
        try:
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                if _type.is_dict_and_not_other(dict, self):
                    self[k] = self.from_dict(v)
                else:
                    self[k] = v
            except:
                self._check_frozen(thrown_error=True)
                raise AttributeError(k)
        else:
            object.__setattr__(self, k, v)

    def __delattr__(self, k):
        self._check_frozen(thrown_error=True)
        try:
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                del self[k]
            except KeyError:
                raise AttributeError(k)
        else:
            object.__delattr__(self, k)

    def copy(self, freeze: bool=False):
        """
        Creaate the new dictionary that is copied this dictionary..
        if pass `freeze=True`, return frozen list object.
        """
        new =  copy.copy(self)
        new.freeze(freeze)
        return new

    def clear(self):
        """ Remove all items from this dictionary."""
        self._check_frozen(thrown_error=True)
        return super().clear()

    def pop(self, key: Hashable, default: Optional[Any]=None):
        """
        remove specified key and return the corresponding value.
        If the key is not found, return the default if given; otherwise,
        raise a KeyError.
        """
        self._check_frozen(thrown_error=True)
        return super().pop(key, default)

    def popitem(self, key: Hashable, value: Optional[Any]=None):
        """
        Remove and return a (key, value) pair as a 2-tuple.
        Pairs are returned in LIFO (last-in, first-out) order.
        Raises KeyError if the dict is empty.
        """
        self._check_frozen(thrown_error=True)
        return super().popitem(key, value)

    def setdefault(self, key, default=None):
        """
        Insert key with a value of default if key is not in the dictionary.
        Return the value for key if key is in the dictionary, else default.
        """
        self._check_frozen(thrown_error=True)
        if key in self:
            return self[key]
        else:
            self[key] = default
            return default

    def freeze(self, shouldFreeze=True):
        """ Freeze this object.  """
        object.__setattr__(self, '__frozen', shouldFreeze)
        for key, val in self.items():
            if _type.is_same_as(val, self):
                val.freeze(shouldFreeze)

    def unfreeze(self):
        """ Unfreeze this object.  """
        self.freeze(False)

    def update(self, *args, **kwargs):
        self._check_frozen(thrown_error=True)
        for key, val in dict(*args, **kwargs).items():
            if _type.is_dict_and_not_other(val, self):
                self[key] = self.from_dict(val)
            else:
                self[key] = val


class uDict(IODict):
    __hash__ = None
    _keypath_separator=Default_Keypath_Separator

    def __init__(self,
            *args: Any,
            separator: str=Default_Keypath_Separator,
            format: Optional[str]=None,
            **kwargs: Any
        ):
        if separator:
            self._keypath_separator = separator
        if format:
            format = str(format).lower()
            if format in ["yml", "yaml"]:
                io.yaml_initializer(cls=uDict)
            args = (super()._decode_init(*args, format=format, **kwargs),)
        else:
            if args and isinstance(args[0], str) and io.validate_file(args[0]):
                format = io.autodetect_format(args[0])
                args = (super()._decode_init(*args, format=format, **kwargs),)
        self.update(dict(*args, **kwargs))

    def __is_keypath_or_keylist(self, x):
        return ( self._keypath_separator is not None
                 and _type.is_keylist_or_keypath(x) )

    def __is_keylist(self, x):
        return ( self._keypath_separator is not None
                 and _type.is_keylist(x) )

    def __is_keypath(self, x):
        return ( self._keypath_separator is not None
                 and _type.is_keypath(x) )

    @property
    def keypath_separator(self):
        return self._keypath_separator

    @keypath_separator.setter
    def keypath_separator(self, val):
        if _type.is_str_not_empty(val):
            self._separator = val
        else:
            raise ValueError('separator must be not empty str.')


    def __missing__(self, key):
        return None

    def __contains__(self,
            key: Union[str, Keylist, Keypath],
        ):
        if self.__is_keypath(key):
            allkeys = self.get_keys(as_keypath=True)
            return key in allkeys
        elif self.__is_keylist(key):
            allkeys = self.get_keys(as_keylist=True)
            return key in allkeys
        elif _type.is_str(key):
            allkeys = self.get_keys()
            return key in allkeys

        return super().__contains__(key)


    def __delitem__(self,
            key: Union[str, Keylist, Keypath],
        ):
        if self.__is_keypath_or_keylist(key):
            self.del_items(key)
            return
        super().__delitem__(key)

    def __getitem__(self,
            key: Union[Hashable, Keylist, Keypath],
        ):
        if ( self.__is_keypath_or_keylist(key)
             or _type.is_tuple(key) ):
            return self.get_values(key)
        return super().__getitem__(key)

    def __setitem__(self,
            key: Union[Hashable, Keylist, Keypath],
            val: Any,
        ):
        if self.__is_keypath_or_keylist(key):
            return self.set_items(key, val)
        super().__setitem__(key, val)

    def __or__(self, other):
        return d.d_merge(self, other,
                         overwrite=True, concat=True,
                         inplace=False, factory=type(self))

    def __ror__(self, other):
        return d.d_merge(other, self,
                         overwrite=True, concat=True,
                         inplace=False, factory=type(self))

    def __ior__(self, other):
        self.update(other)
        return self

    def __str__(self):
        return f'{self.to_dict(self)}'

    def get(self,
            key: Union[str, Keylist, Keypath],
            default=None,
        ):
        if self.__is_keypath_or_keylist(key):
            return self.get_values(key)
        return super().get(key, default)

    def pop(self,
            key: Union[str, Keylist, Keypath],
            *args: Any,
        ):

        default=args[0] if args else None
        if self.__is_keypath_or_keylist(key):
            return self.pop_items(key, default=default)
        else:
            return super().pop(key, default)

    def set(self,
            key: Union[str, Keylist, Keypath],
            val: Any,
        ):
        self[key] = val

    def setdefault(self,
            key: Union[str, Keylist, Keypath],
            default=None,
        ):
        if key not in self:
            self[key] = default
            return default
        return self[key]


    @staticmethod
    def _get_docstring(obj, attr: Optional[str]=None):
        if attr:
            doc = f"{obj.__doc__}\nIf {attr} is omitted, self is used."
        else:
            doc = f"{obj.__doc__}"
        return doc

    def clean(self,
            obj: Optional[dict]=None,
            strings=True,
            collections=True,
            *,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new =  d.d_clean(obj, strings, collections,
                                       inplace=False, factory=dict)
        if inplace:
            self.clear()
            self.update(new)
        else:
            return factory(new)

    clean.__doc__ = _get_docstring(d.d_clean, 'obj')

    def clone(self,
            obj: Optional[dict]=None,
            empty: bool=False,
            memo: Optional[dict]=None,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new = d.d_clone(obj, empty, memo)
        return factory(new)

    clone.__doc__ = _get_docstring(d.d_clone, 'obj')

    def compare(self,
        d1: dict,
        d2: Optional[dict]=None,
        *,
        keys: Optional[Union[Hashable,list, Keylist, Keypath]]=None,
        thrown_error: bool=False,
        ):
        d2 = d2 or self
        return d.d_compare(d1, d2, keys=keys, thrown_error=thrown_error)

    compare.__doc__ = _get_docstring(d.d_compare, 'd2')

    def counts(self,
            pattern: Union[Pattern, Hashable, Sequence],
            obj: Optional[dict]=None,
            *,
            count_for: DictItemType=DictItem.KEY,
            wild: bool=False,
            verbatim: bool=False,
        ) ->Union[int, dict]:

        obj = obj if obj or obj == {} else self
        return d.d_counts(obj, pattern, count_for=count_for,
                        wild=wild, verbatim=verbatim)

    counts.__doc__ = _get_docstring(d.d_counts, 'obj')


    def filter(self,
            predicate: Callable,
            obj: Optional[dict]=None,
            *,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new = factory()

        return d.d_filter(predicate, obj, factory=factory)

    filter.__doc__ = _get_docstring(d.d_filter, 'obj')

    def groupby( self,
            seq: list,
            key: Hashable,
            *,
            factory: Optional[Type[dict]]=None,
        ) -> dict:
        factory = factory or type(self)
        return d.d_groupby(seq, key, factory=factory)

    groupby.__doc__ = _get_docstring(d.d_groupby)

    def invert( self,
            obj: Optional[dict]=None,
            flat: bool=False,
            *,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) ->dict:
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        return d.d_invert(obj, flat=flat,
                               inplace=inplace, factory=factory)

    invert.__doc__ = _get_docstring(d.d_invert, 'obj')

    def map(self,
            func: Callable,
            obj: Optional[dict]=None,
            *,
            map_for: Optional[DictItemType]=None,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) ->dict:

        obj = obj if obj or obj == {} else self
        factory = factory or type(self)

        return d.d_map(func, obj, map_for=map_for,
                       inplace=inplace, factory=factory)

    map.__doc__ = _get_docstring(d.d_map, 'obj')

    def merge(self,
            others: list,
            obj: Optional[dict]=None,
            *,
            overwrite: bool=True,
            concat: bool=False,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) ->dict:

        obj = obj if obj or obj == {} else self
        factory = factory or type(self)

        return d.d_merge(obj, others,
                         overwrite=overwrite, concat=concat,
                         inplace=inplace, factory=factory)

    merge.__doc__ = _get_docstring(d.d_merge, 'obj')

    def move(self,
            key_src: Union[str, list],
            key_dest: Union[str, list],
            obj: Optional[dict]=None,
            *,
            overwrite: bool=True,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) ->dict:
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new = d.d_move(obj, key_src, key_dest,
                        overwrite=overwrite,
                        inplace=False, factory=dict)
        return d._dict_updator(obj, new, inplace=inplace, factory=factory)

    move.__doc__ = _get_docstring(d.d_move, 'obj')


    def nest(self,
        items: tuple,
        id_key: Union[str, list],
        parent_id_key: Union[str, list],
        children_key: Union[str, list],
        )->list:

       return d.d_nest(items, id_key, parent_id_key, children_key)

    nest.__doc__ = _get_docstring(d.d_nest)


    def remove(self,
            keys: Union[list, Hashable],
            obj: Optional[dict]=None,
            *,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        return d.d_remove(obj, keys,
                      inplace=inplace, factory=factory)

    remove.__doc__ = _get_docstring(d.d_remove, 'obj')

    def rename(self,
            key: Union[Hashable,dict],
            key_new: Optional[Hashable]=None,
            obj: Optional[dict]=None,
            case_name: Optional[str]=None,
            *,
            overwrite: bool=False,
            keep_order: bool=False,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) ->dict:
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new = d.d_rename(obj, key, key_new,
                        case_name=case_name, overwrite=overwrite,
                        inplace=False, factory=dict)
        return d._dict_updator(obj, new, inplace=inplace, factory=factory)


    rename.__doc__ = _get_docstring(d.d_rename, 'obj')


    def get_keys(self,
            obj: Optional[dict]=None,
            *,
            indexes: bool=False,
            output_as: Optional[DictKeyType]=None,
            separator: str=Default_Keypath_Separator,
        ) -> list:
        obj = obj if obj or obj == {} else self
        return d.get_keys(obj, indexes=indexes,
                               output_as=output_as,
                               separator=separator)

    get_keys.__doc__ = _get_docstring(d.get_keys, 'obj')

    def get_values(self,
            keys: Union[Hashable, Sequence],
            obj: Optional[Union[dict, Sequence]] = None,
        ) -> Union[list, dict]:
        obj = obj if obj or obj == {} else self
        return d.get_values(obj, keys)

    get_values.__doc__ = _get_docstring(d.get_values, 'obj')

    def get_items(self,
            loc: Union[Hashable, Keylist, Keypath],
            value: Optional[Any]=None,
            obj: Optional[dict]=None,
            *,
            func: Optional[Callable]=None,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new = factory(obj)
        new = d.get_items(new, loc, value, func=func, factory=factory)
        return new

    get_items.__doc__ = _get_docstring(d.get_items, 'obj')

    def pop_items(self,
            loc: Hashable,
            value: Optional[Any]=None,
            obj: Optional[dict]=None,
            *,
            func: Optional[Callable]=None,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        item = d.pop_items(obj, loc, value, func=func, factory=factory)
        return item

    pop_items.__doc__ = _get_docstring(d.pop_items, 'obj')

    def del_items(self,
            loc: Union[Hashable, Keylist, Keypath],
            obj: Optional[dict]=None,
            *,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)

        return d.del_items(obj, loc, inplace=inplace, factory=dict)

    del_items.__doc__ = _get_docstring(d.del_items, 'obj')

    def set_items( self,
            loc: Union[str, Sequence],
            value: Any,
            obj: Optional[Union[dict, Sequence]]=None,
            func: Optional[Callable]=None,
            *,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)

        d.set_items(obj, loc, value, func=func, factory=factory)

    set_items.__doc__ = _get_docstring(d.set_items, 'obj')

    def find( self,
            keys: Union[list,Hashable],
            default: Optional[Any]=None,
            obj: Optional[dict]=None,
            *,
            first_one: bool=True
        ):
        obj = obj if obj or obj == {} else self
        return d.d_find(obj, keys, default, first_one)


    find.__doc__ = _get_docstring(d.d_find, 'obj')

    def flatten( self,
            obj: Optional[dict]=None,
            *,
            separator: str=Default_Keypath_Separator,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) -> dict:
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new =  d.d_flatten(obj, separator, inplace=False, factory=factory)
        return d._dict_updator(obj, new, inplace=inplace, factory=factory)

    flatten.__doc__ = _get_docstring(d.d_flatten, 'obj')

    def unflatten( self,
            obj: Optional[dict]=None,
            default: Optional[Any]=None,
            *,
            separator: str=Default_Keypath_Separator,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ) -> dict:
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        new =  d.d_unflatten(obj, default, separator,
                                  inplace=False, factory=factory)
        return d._dict_updator(obj, new, inplace=inplace, factory=factory)

    unflatten.__doc__ = _get_docstring(d.d_unflatten, 'obj')


    def keylists( self,
            obj: Optional[dict]=None,
            indexes: bool=False,
        ) -> list:
        obj = obj if obj or obj == {} else self
        keylists = []
        for k in d.keylists(obj, indexes):
            keylists.append(Keylist(k, separator=self.keypath_separator))
        return keylists


    keylists.__doc__ = _get_docstring(d.keylists, 'obj')

    def keypaths( self,
            obj: Optional[dict]=None,
            indexes: bool=False,
            *,
            separator: Optional[str]=None,
        ) -> list:
        obj = obj if obj or obj == {} else self
        separator = separator or self.keypath_separator
        kps = []
        for kp in d.keypaths(obj, indexes=indexes, separator=separator):
            kps.append(Keypath(kp, separator=separator))
        return kps

    keypaths.__doc__ = _get_docstring(d.keypaths, 'obj')

    def search( self,
            query: Pattern,
            obj: Optional[dict]=None,
            *,
            search_for: DictItemType=DictItem.KEY,
            exact: bool=False,
            ignore_case: bool=False,
            use_keypath: bool=True,
        ):
        obj = obj if obj or obj == {} else self
        return d.d_search(obj, query,
                          search_for=search_for,
                          exact=exact,
                          ignore_case=ignore_case)

    search.__doc__ = _get_docstring(d.d_search, 'obj')

    def sort( self,
            obj: Optional[dict]=None,
            *,
            sort_by: DictItemType=DictItem.KEY,
            reverse: bool=False,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        return d.d_sort(obj, sort_by, reverse=reverse,
                           inplace=inplace, factory=factory)
    sort.__doc__ = _get_docstring(d.d_sort, 'obj')


    def subset( self,
            keys: Union[str, list, tuple, Hashable],
            default: Optional[Any]=None,
            obj: Optional[dict]=None,
            *,
            use_keypath: bool=False,
            separator: Optional[str]=None,
            inplace: bool=False,
            factory: Optional[Type[dict]]=None,
        ):
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        separator = separator or self.keypath_separator
        return d.d_subset(obj, keys, default=default,
                        use_keypath=use_keypath,
                        separator=separator,
                        inplace=inplace, factory=factory)

    subset.__doc__ = _get_docstring(d.d_subset, 'obj')

    def swap( self,
            key1: Hashable,
            key2: Hashable,
            obj: Optional[dict]=None,
            *,
            inplace: bool=False,
            factory: Optional[dict]=None
        ) ->Optional[dict]:
        obj = obj if obj or obj == {} else self
        factory = factory or type(self)
        return d.d_swap(obj, key1, key2,
                        inplace=inplace, factory=factory)

    swap.__doc__ = _get_docstring(d.d_swap, 'obj')


    def traverse( self,
            callback: Callable,
            obj: Optional[Union[dict, list, tuple]]=None,
            parents: list=[],
            *args: Any,
            **kwargs: Any,
        ):
        obj = obj if obj or obj == {} else self
        return d.d_traverse(obj, callback, parents, *args, **kwargs)

    traverse.__doc__ = _get_docstring(d.d_traverse, 'obj')


    def unique( self,
            obj: Optional[dict]=None,
        ) -> list:
        obj = obj if obj or obj == {} else self
        return d.d_unique(obj)

    unique.__doc__ = _get_docstring(d.d_unique, 'obj')

    def update(self, *args, **kwargs):
        for key, val in dict(*args, **kwargs).items():
            if _type.is_dict_and_not_other(val, self):
                self[key] = self.from_dict(val)
            else:
                self[key] = val



class iList(list):

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, '_attrs', aDict())

    @property
    def attrs(self):
        return self._attrs

    def get_attrs(self, name: Optional[str]=None, default=None):
        if not name:
            return self._attrs.to_dict()
        else:
            return self._attrs.get(name, default)

    def _check_frozen(self,
            thrown_error: bool=False,
            msg: str='frozen object cannot be modified.',
            ):
        if self._attrs._check_frozen(thrown_error=False):
            if thrown_error:
                raise AttributeError( f"{self.__class__.__name__} {msg}" )
            else:
                return True
        else:
            return False

    def __add__(self, other): # self + other
        """
        Create new list that is Concatenate this list with another list.
        """
        self._check_frozen(thrown_error=True)
        return super().__add__(other)

    def __radd__(self, other): # self + other
        """
        Concatenate this list with another list
        """
        self._check_frozen(thrown_error=True)
        return super().__add__(other)

    def __iadd__(self, other):  # self += other
        """
        Concatenate this list with another list in-placed.
        """
        self._check_frozen(thrown_error=True)
        return super().__iadd__(other)

    def __imul__(self, other):  # self *= other
        """
        Return value*self.
        """
        self._check_frozen(thrown_error=True)
        return super().__imul__(other)

    def __and__(self, other):  # self & other
        self._check_frozen(thrown_error=True)
        return list(set(self).intersection(other))

    def __rand__(self, other):  # self & other
        self._check_frozen(thrown_error=True)
        return list(set(other).intersection(self))

    def __or__(self, other):  # self | other
        new = list(set([*self, *other]))
        return new

    def __xor__(self, other):  # self ^ other
        new = list(set(self).symmetric_difference(set(other)))
        return new

    def __sub__(self, other): # self - other
        return list(set(self).difference(other))

    def __rsub__(self, other): # self - other
        return list(set(other).difference(self))

    def __setitem__(self, index, val):
        self._check_frozen(thrown_error=True)
        return super().__setitem__(index, val)

    def __setattr__(self, name, val):
        self._check_frozen(thrown_error=True)
        self._attrs[name] = val

    def __getattr__(self, item):
        if item in self._attrs:
            return self._attrs[item]
        else:
            self.__missing__(item)

    def __missing__(self, name):
        raise KeyError(f'{name} not found.')

    def __delattr__(self, name):
        self._check_frozen(thrown_error=True)
        del self._attrs[name]

    def __hash__(self):
        if self._check_frozen(thrown_error=False):
            return hash(frozenset(self.__repr__()))
        else:
            raise AttributeError('unhashable not frozen object.')

    def __reversed__(self):
        """ Create new list revesesed items.  """
        for element in self[::-1]:
            yield element

    def __repr__(self, verbose: bool=False):
        str_value = str(list(self))
        return f'{self.__class__.__name__}({str_value})'

    def append(self, other):
        """
        Append object to the end of the list.
        """
        self._check_frozen(thrown_error=True)
        super().append(other)

    def reverse(self):
        """ Reverse *IN PLACE* """
        self._check_frozen(thrown_error=True)
        super().reverse()

    def clear(self):
        """
        Remove all items from list.
        """
        self._check_frozen(thrown_error=True)
        super().clear()

    def copy(self, freeze: bool=False):
        """
        Creaate the new list that is copied this list.
        this method could not copy self.attrs..
        if pass `freeze=True`, return frozen list object.
        """

        new = type(self)([x for x in self])
        new.freeze(freeze)
        return new

    def clone(self, empty: bool=False):
        """
        Creaate the new list that is cloned this list.
        this method copy self.attrs.
        if pass `empty=True`, keep self.attrs but list will be cleared.
        """

        if empty:
            new = type(self)([])
        else:
            new = type(self)([x for x in self])

        new.attrs.from_dict(self.attrs.to_dict(), inplace=True)
        new.attrs.freeze(self.attrs._check_frozen(thrown_error=False))
        return new

    def extend(self, obj: Iterable):
        """
        Extend list by appending elements from the iterable.
        """
        self._check_frozen(thrown_error=True)
        super().extend(obj)

    def insert(self, index: inte, obj: Iterable):
        """
        Insert object before index
        """
        self._check_frozen(thrown_error=True)
        super().insert(index, obj)

    def pop(self, index: int=-1):
        """
        Remove and return item at index (default last).
        Raises IndexError if list is empty or index is out of range.
        """
        self._check_frozen(thrown_error=True)
        return super().pop(index)

    def remove(self, val: Any):
        """
        Remove first occurrence of value.
        Raises ValueError if the value is not present.
        """
        self._check_frozen(thrown_error=True)
        return super().remove(val)

    def without(self,
            *items: Any,
        ) ->list:
        """ Create new list without items and return iterable.  """
        return list(set(self).difference(*items))

    def find(self,
            val: Union[Any, list, tuple],
        ) -> list:
        """
        Return the list of index that found val in list.
        otherwise return None
        """
        if _type.is_list_or_tuple(val):
            found = [ i for v in val for i, x in enumerate(self) if x == v ]
        else:
            found = [ i for i, x in enumerate(self) if x == val ]

        return found if found else None

    def replace(self,
            old: Any,
            new: Any,
            func: Optional[Callable]=None,
        ) ->list:
        """
        Return a new list that has new instead of old.
        if old is not found, it will raise an ItemNotFountError.
        callback function will be called as follows.
            `func(index, old, new)`
        """
        indexes = self.find(old)
        if indexes is None:
            raise ValueError(f'Item {old} not found in list.')

        func = func if func else lambda x, y, z: new
        new_list = list(self)
        for i in indexes:
            new_list[i] = func(i, old, new)

        return new_list

    def sort(self, key: Optional[Callable]=None, reverse: bool=False):
        """
        Sort the list in ascending order and return None.
        The sort is in-place (i.e. the list itself is modified)
        and stable (i.e. the order of two equal elements is maintained).
        If a key function is given,
        apply it once to each list item and sort them,
        ascending or descending, according to their function values.
        The reverse flag can be set to sort in descending order.
        """
        self._check_frozen(thrown_error=True)
        super().sort(key=key, reverse=reverse)

    def freeze(self, shouldFreeze=True):
        self._attrs.freeze(shouldFreeze)

    def unfreeze(self):
        self._attrs.unfreeze()

    def update(self, *args, **kwargs):
        self._check_frozen(thrown_error=True)
        self.__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

import re
from typing import Any, Dict, Union, Optional, Hashable
from collections.abc import Mapping
from functools import partial
from unicodedata import normalize
#
import numpy as np
from multimethod import multidispatch, multimethod
from datajuggler.validator import (
    DictItem, DictItemType, validate_DictItem, validate_DictItem
)

from datajuggler.strings import substr
from datajuggler.validator import TypeValidator as _type

try:
    import pandas as pd
    pd_NA = pd.NA
    pandas_installed = True
except ImportError:
    pd_NA = None
    pandas_installed = False


class StrCase(object):

    def __init__(self, *args: Any):
        self.depth = 0
        self.__NULL_VALUES = {"", None, np.nan, pd_NA}

        self.supported_case = {
            "snake": {
                'sample': 'convert_case',
                'separaator': '_',
                'splitor': self.split_strip_string,
                'convertor': lambda x: "_".join(x).lower() },
            "kebab": {
                'sample': 'convert-case',
                'separaator': '-',
                'splitor': self.split_strip_string,
                'convertor': lambda x: "-".join(x).lower() },
            "camel": {
                'sample': 'convertCase',
                'separaator': '',
                'splitor': self.split_strip_string,
                'convertor': lambda x: x[0].lower() + "".join(w.capitalize() for w in x[1:]) },
            "pascal": {
                'sample': 'ConvertCase',
                'separaator': '',
                'splitor': self.split_strip_string,
                'convertor': lambda x:  "".join(w.capitalize() for w in x) },
            "const": {
                'sample': 'CONVERT_CASE',
                'separaator': '_',
                'splitor': self.split_strip_string,
                'convertor':  lambda x: "_".join(x).upper() },
            "sentence": {
                'sample': 'Convert case',
                'splitor': self.split_string,
                'convertor':  lambda x: " ".join(x).capitalize() },
            "title": {
                'sample': 'Convert Case',
                'separaator': ' ',
                'splitor': self.split_string,
                'convertor':  lambda x: " ".join(w.capitalize() for w in x) },
            "lower": {
                'sample': 'convert case',
                'separaator': ' ',
                'splitor': self.split_string,
                'convertor':  lambda x:  " ".join(x).lower() },
            "upper": {
                'sample': 'CONVERT CASE',
                'separaator': ' ',
                'splitor': self.split_string,
                 'convertor': lambda x: " ".join(x).upper() },
        }

        if len(args) == 0:
            self.__origin = None
        elif len(args) == 1:
            self.__origin = self.__validate(args[0])
        else:
            raise TypeError(
                'Expected at most 1 arguments, got {}.'.format(len(args)))

    def __validate(self,
            val: Union[str, list, dict]
        )-> Union[str, list, dict]:

        if _type.is_str(val) or _type.is_list(val) or _type.is_dict(val):
            return val
        else:
            raise TypeError( 'Expected str or list, dict objects, got {}.'
                             .format(type(val)) )

    @property
    def origin(self):
        return self.__origin

    @origin.setter
    def origin(self, val):
        self.__origin = self.__validate(val)

    def show_supported_case(self, verbose=False):
        header = { "case":  "sample" }
        case_sample = dict(header,
           **{ case: "{}".format(self.supported_case[case]['sample'])
               for case in self.supported_case.keys() } )
        if verbose:
            print(case_sample)
        return case_sample

    @classmethod
    def split_strip_string(cls, string: str) -> list:
        """Split the string into separate words and strip punctuation."""
        string = re.sub(r"[!()*+\,\-./:;<=>?[\]^_{|}~]", " ", string)
        string = re.sub(r"[\'\"\`]", "", string)

        return re.sub(
            r"([A-Z][a-z]+)",
            r" \1", re.sub(r"([A-Z]+|[0-9]+|\W+)",
            r" \1",
            string)
        ).split()

    @classmethod
    def split_string(cls, string: str) -> list:
        """Split the string into separate words."""
        string = re.sub(r"[\-_]", " ", string)

        return re.sub(
            r"([A-Z][a-z]+)",
            r" \1",
            re.sub(r"([A-Z]+)",
            r"\1", string)
        ).split()

    def __convert_case_str(self,
            case: str,
            data: str,
        )-> str:

        if data != '':
            if case in self.supported_case:
                words = self.supported_case[ case ]['splitor'](data)
                data = self.supported_case[ case ]['convertor'](words)
            else:
                raise ValueError(
                    'Invalid casename, '
                    'check class.show_supported_case().' )

        return data

    def __convert_case_dict(self,
            case: str,
            data: dict,
            replace_for: DictItemType
           ) -> dict:

        self.depth += 1
        if replace_for == DictItem.KEY:
            convdict = { self.convert_case(case, x, replace_for ):y
                         for x, y in data.items() }
        elif replace_for == DictItem.VALUE:
            convdict = { x:self.convert_case(case, y, replace_for )
                         for x, y in data.items() }
        self.depth -= 1
        return convdict

    def convert_case(self,
            case: str='snake',
            data: Optional[Union[list,str, dict]] = None,
            replace_for: DictItemType = DictItem.VALUE
        ) -> str:
        """Convert case style for obj.

        Parameters
        ----------
        case: str
            Preferred case type, i.e.:  (default: 'snake')
            check `.show_supported_case()`

        obj: Any
            convert case for data

        Returns:
        --------
        converted data: str
        """

        validate_DictItem(replace_for, thrown_error=True)

        if data is None:
            if self.depth == 0:
                if self.__origin:
                    data = self.__origin
                else:
                    raise ValueError('Expected at most 1 objets')
            else:
                return data

        if _type.is_str(data):
            return self.__convert_case_str(case, data)
        elif _type.is_list(data):
            self.depth += 1
            convobj = [ self.convert_case(case, x, replace_for)
                        for x in data ]
            self.depth -= 1
            return convobj
        elif _type.is_dict(data):
            return self.__convert_case_dict(case, data, replace_for)
        else:
            return data

    def __str__(self) -> str:
        return str(self.origin)

    def __repr__(self) -> str:
        if _type.is_str(self.__origin):
            return 'StrCase("{}")'.format(self.__origin)
        else:
            return 'StrCase({})'.format(self.__origin)

class urange(object):
    def __init__(self,
            *args: int,
            **kwargs: Any
        ):
        """Return an object that produces a sequence of integes

        range(stop) -> range object
        range(start, stop[, step]) -> range object
        range(start, stop[, func]) -> range object

        from start (inclusive) to stop (exclusive) by step.

          range(i, j) produces i, i+1, i+2, ..., j-1.
          range(i, j, k) produces i, i+k, i+2, ..., j-k.

        start defaults to 0, and stop is omitted!

          range(4) produces 0, 1, 2, 3.

        These are exactly the valid indices for a list of 4 elements.
        When step is given, it specifies the increment (or decrement).

        if set step as callable, pass parameter current value.
        So, you should define as following function.
          def step(val: int)->int:
              # something to do for val
              return val
        """

        if len(args) == 1:
            self._i = 0
            self._end = args[0]
            self._step = kwargs.get('step', 0)
        elif len(args) == 2:
            self._i = args[0]
            self._end = args[1]
            self._step = kwargs.get('step', 0)
        elif len(args) >= 3:
            self._i = args[0]
            self._end = args[1]
            self._step = args[2]
        else:
            self._i = kwargs.get('start', 0)
            self._end = kwargs.get('end')
            self._step = kwargs.get('step', 0)

        self.step_callable = callable(self._step)
        if self._i <= self._end:
            self.ascending = True
            if self._step == 0:
                self._step = 1
            elif not self.step_callable and (self._step < 0):
                raise ValueError(
                      'Step must be postitive value for ascending orders.')
        else:
            self.ascending = False
            if self._step == 0:
                self._step = -1
            elif not self.step_callable and (self._step > 0):
                raise ValueError(
                      'Step must be negative value for descending orders.')

    def __iter__(self):
        return self

    def __next__(self):
        if self.ascending:
            if self._i >= self._end:
                raise StopIteration
        else:
            if self._i <= self._end:
                raise StopIteration

        value = self._i
        if self.step_callable:
            self._i = self._i + self._step(self._i)
        else:
            self._i = self._i + self._step
        return value

def rename_duplicates(
        data: list,
        separator: str = '_',
        format = "{:02}"
    ) -> list:
    """Rename duplicated strings to append a number at the end."""

    counts: Dict[str, int] = {}

    if _type.is_list(data):
        for i, val in enumerate(data):
            if _type.is_list(val):
                new_val = rename_duplicates(val)
                data[i] = new_val
            elif _type.is_str(val):
                cur_count = counts.get(val, 0)
                if cur_count > 0:
                    data[i] = ( "{}{}".format(val, separator)
                                + format.format(cur_count) )
                counts[val] = cur_count + 1
    return data

@multidispatch
def replace_values( obj: Any, *arg: Any, **kwargs: Any) ->Any:
    """dispatch function replace_values """
    if obj:
        return obj

@replace_values.register(list, dict)
def _replace_values_multi(
        values: list,
        replace: dict,
        *,
        ignore_case: bool=False,
        inplace: bool=False,
        **kwargs: Any,
    )-> Optional[list]:

   if not inplace:
       values = values.copy()

   for n in range(len(values)):
       for old, new in replace.items():
           values[n] = substr(old, new,  values[n], ignore_case=ignore_case)
   if not inplace:
       return values

@replace_values.register(dict, dict)
def _replace_values_dict_multi(
        values: dict,
        replace: dict,
        *,
        ignore_case: bool=False,
        inplace: bool=False,
        replace_for: DictItemType = DictItem.VALUE
    )-> Optional[list]:
    """replace values of dict
    Parameters
    ----------
    values: dict
       to replace data
    replace: dict
       replace map. { old: new,....}
    replace_for: avaiable 'key', 'value'.
       If replace_for set 'value', replace 'value' of dict.
       If replace_value set 'key', replace 'key' of dict.
       default is 'value'
    """

    def replace_val(value, replace, ignore_case):
       for old, new in replace.items():
           if ignore_case :
               if value.lower() == old.lower():
                   return new
               else:
                   value
           else:
               if value == old:
                   return new
               else:
                   value
       return value

    validate_DictItem(replace_for, thrown_error=True)

    mapper={}
    if replace_for == DictItem.KEY:
       keys = [ replace_val(x, replace, ignore_case)
                for x in values.keys() ]
       vals = list(values.values())
    elif replace_for == DictItem.VALUE:
       keys = list(values.keys())
       vals = [replace_val(x, replace, ignore_case)
               for x in values.values() ]
    else:
       return None

    workdict = dict(zip(keys, vals))

    if inplace:
        values.update(workdict)
    else:
        return workdict

@replace_values.register(list, list, str)
def _replace_values_single_str(
        values: list,
        replace_from: list,
        replace_to: str,
        *,
        ignore_case: bool=False,
        inplace: bool=False,
        **kwargs: Any,
    )-> Optional[list]:

   if not inplace:
       values = values.copy()

   for n in range(len(values)):
       for old in replace_from:
           values[n] = substr( old, replace_to, values[n],
                                    ignore_case=ignore_case)
   if not inplace:
       return values

@replace_values.register(list, list, Any)
def _replace_values_single_obj(
        values: list,
        replace_from: list,
        replace_to: Any,
        *,
        ignore_case: bool=False,
        inplace: bool=False,
        **kwargs: Any,
    )-> Optional[list]:

   if not inplace:
       values = values.copy()

   for n in range(len(values)):
       for old in replace_from:
           if _type.is_str(values[n]) and ignore_case:
               new = replace_values(values[n], [old], replace_to)
           if values[n] ==  old:
               values[n] = new

   if not inplace:
       return values


@replace_values.register(str, list, Hashable)
def _replace_values_text(
        values: str,
        replace_from: list,
        replace_to: Hashable,
        *,
        ignore_case: bool=False,
        **kwargs: Any,
    )-> Hashable:

   for old in replace_from:
       if _type.is_str(old):
           if not _type.is_str(replace_to):
               replace_to = str(replace_to)
           values = substr(old, replace_to, values, ignore_case=ignore_case)
       else:
           if values == old:
               values = replace_to

   return values

@replace_values.register(str, dict)
def _replace_values_text_multi(
        values: str,
        replace: dict,
        *,
        ignore_case: bool=False,
        **kwargs: Any,
    )-> Hashable:

   for old, replace_to in replace.items():
       if _type.is_str(old):
           if not _type.is_str(replace_to):
               replace_to = str(replace_to)
           values = substr(old, replace_to, values,
                                ignore_case =ignore_case)
       else:
           if values == old:
               values = replace_to

   return values

@replace_values.register(Union[int, float], list, Any)
def _replace_values_number(
        values: Union[int, float],
        replace_from: list,
        replace_to: Any,
        **kwargs: Any,
    )-> str:
   if values in replace_from:
       return replace_to
   else:
       return values


@multidispatch
def omit_values( obj: Any, *arg: Any, **kwargs: Any) ->Any:
    """dispatch function replace_values """
    if obj:
        return obj

@omit_values.register(list, list)
def _omit_values_multi(
        values: list,
        omits: list,
        *,
        inplace: bool=False,
        ignore_case: bool=False,
        drop: bool=False
    )-> list:

   if not inplace:
       values = values.copy()

   values =  replace_values(values, omits, '', ignore_case=ignore_case)
   if drop:
       count = values.count('')
       for _ in range(count):
           values.remove('')

   if not inplace:
       return values

@omit_values.register(str, list)
def omit_values_single(
        values: str,
        omits: list,
        *,
        ignore_case: bool=False,
    )-> str:
    return replace_values(values, omits, '', ignore_case=ignore_case)

@multidispatch
def split_chunks( *args: Any, **kwargs: Any ):
    """ dispatch function split_chunks """
    raise TypeError('Invalid Type')

@split_chunks.register
def _split_chunks_list(
        data: list,
        chunk_size: int,
        fill_na: bool=True,
        na_value: Optional[Union[int,str,float]]=None
    ) ->list:
    """Return split into even chunk_size elements for iterable
    Parameters
    ----------
    data: list
        to split itrtable ojects. i.e.: list, tuple, dict, set.
    chunk_size: int
        the size of chunk list.
    fill_na: bool
        subset of a list does not fit in the size of the defined chunk,
        fillers need to be inserted in the place of the empty element holders.
        fillers is able to set `na_value`.
    na_value: int, str, default is `None`.

    Returns
    -------
        splited list
    """

    iterable = iter(data)
    for x in range(0, len(data), chunk_size):
        chunk_data = list()
        for _ in urange(chunk_size):
            try:
                chunk_data.append(next(iterable))
            except StopIteration:
                break

        if fill_na and (len(chunk_data) < chunk_size):
            chunk_data += [ na_value
                            for y in urange(chunk_size-len(chunk_data))]
        yield chunk_data

@split_chunks.register
def _split_chunks_tuple(
        data: tuple,
        chunk_size: int,
        fill_na: bool=True,
        na_value: Optional[Union[int,str,float]]=None
    ) ->list:
    """Return split into even chunk_size elements for iterable
    Parameters
    ----------
    data: tuple
        to split itrtable ojects. i.e.: list, tuple, dict, set.
    chunk_size: int
        the size of chunk list.
    fill_na: bool
        subset of a list does not fit in the size of the defined chunk,
        fillers need to be inserted in the place of the empty element holders.
        fillers is able to set `na_value`.
    na_value: int, str, default is `None`.

    Returns
    -------
        splited list
    """

    iterable = iter(data)
    for x in range(0, len(data), chunk_size):
        chunk_data = list()
        for _ in urange(chunk_size):
            try:
                chunk_data.append(next(iterable))
            except StopIteration:
                break

        if fill_na and (len(chunk_data) < chunk_size):
            chunk_data += [ na_value
                            for y in range(chunk_size-len(chunk_data))]
        yield tuple(chunk_data)


@split_chunks.register
def _split_chunks_dict(
        data: dict,
        chunk_size: int,
        *args: Any,
        **kwargs: Any,
    ) ->list:
    """Return split into even chunk_size elements for iterable
    Parameters
    ----------
    data: dict
        to split dict ojects.
    chunk_size: int
        the size of chunk list.
    na_value: int, str, default is `None`.

    Returns
    -------
        splited list
    """

    iterable = iter(data)
    for x in range(0, len(data), chunk_size):
        chunk_data = dict()
        for _ in urange(chunk_size):
            try:
                key = next(iterable)
                chunk_data[key] = data[key]
            except StopIteration:
                break

        yield chunk_data

@split_chunks.register
def _split_chunks_str(
        data: str,
        chunk_size: int,
        *args: Any,
        **kwargs: Any,
    ) ->list:
    """Return split into even chunk_size elements for iterable
    Parameters
    ----------
    data: str
        to split dict ojects.
    chunk_size: int
        the size of chunk list.
    na_value: int, str, default is `None`.

    Returns
    -------
        splited str list
    """

    iterable = iter(data)
    for x in range(0, len(data), chunk_size):
        chunk_data = ""
        for _ in urange(chunk_size):
            try:
                 chunk_data += next(iterable).replace('\n', '')
            except StopIteration:
                break

        yield chunk_data


def parse_slice(expr: str):
    """ Parse slice expression and return tupele. """
    def to_piece(s):
        try:
            return s and int(s) or None
        except:
            return None
    pieces = list(map(to_piece, expr.split(':')))
    if len(pieces) == 1:
        return slice(pieces[0], pieces[0] + 1)
    else:
        return slice(*pieces)

def add_df(
        values: list,
        columns: list,
        omits: list=[]
    ) ->pd.DataFrame:
    """Add values to dataframe"""

    if not pandas_installed:
        raise ModuleNotInstalledError("'pandas' module is not installed.")

    if omits:
        values = omit_values(values,omits)
        columns = omit_values(columns,omits)

    # Since Pandas 1.3.0
    df = pd.DataFrame(values,index=columns)._maybe_depup_names(columns)
    self.df = pd.concat([self.df,df.T])

def df_compare(
        df1: pd.DataFrame,
        df2: pd.DataFrame,
    ) -> int:
    """ Compare DataFrame
    Parameters
    ----------
    df1: pd.DataFrame, df2: pd.DataFrame
        any DataFrame to compare

    Returns
    -------
    validate result: Union[bool,int]
    """

    if not pandas_installed:
        raise ModuleNotInstalledError("'pandas' module is not installed.")

    diff_df = pd.concat([df1,df2]).drop_duplicates(keep=False)
    diffs = len(diff_df)
    return diffs




# -*- coding: utf-8 -*-

import re

from pathlib import Path
from typing import Union, Any

try:
    import requests
    requests_installed = True
except ImportError:
    requests_installed = False

from datajuggler.serializer.abstract import AbstractSerializer
from datajuggler.serializer.base64 import Base64Serializer
from datajuggler.serializer.csv import CSVSerializer
from datajuggler.serializer.ini import INISerializer
from datajuggler.serializer.json import JSONSerializer
from datajuggler.serializer.pickle import PickleSerializer
from datajuggler.serializer.plist import PListSerializer
from datajuggler.serializer.query_string import QueryStringSerializer
from datajuggler.serializer.toml import TOMLSerializer
from datajuggler.serializer.xml import XMLSerializer
from datajuggler.serializer.yaml import YAMLSerializer, yaml_initializer


__all__ = [
    "AbstractSerializer",
    "Base64Serializer",
    "CSVSerializer",
    "INISerializer",
    "JSONSerializer",
    "PickleSerializer",
    "PListSerializer",
    "QueryStringSerializer",
    "TOMLSerializer",
    "XMLSerializer",
    "YAMLSerializer",
    "yaml_initializer",
    "get_format_by_path",
    "autodetect_format",
    "validate_file",
    "is_url",
    "read_contents",
    "read_url",
    "read_file",
    "write_file",
]

_BASE64_SERIALIZER = Base64Serializer()
_CSV_SERIALIZER = CSVSerializer()
_INI_SERIALIZER = INISerializer()
_JSON_SERIALIZER = JSONSerializer()
_PICKLE_SERIALIZER = PickleSerializer()
_PLIST_SERIALIZER = PListSerializer()
_QUERY_STRING_SERIALIZER = QueryStringSerializer()
_TOML_SERIALIZER = TOMLSerializer()
_YAML_SERIALIZER = YAMLSerializer()
_XML_SERIALIZER = XMLSerializer()

_SERIALIZERS = {
    "b64": _BASE64_SERIALIZER,
    "base64": _BASE64_SERIALIZER,
    "csv": _CSV_SERIALIZER,
    "ini": _INI_SERIALIZER,
    "json": _JSON_SERIALIZER,
    "pickle": _PICKLE_SERIALIZER,
    "plist": _PLIST_SERIALIZER,
    "qs": _QUERY_STRING_SERIALIZER,
    "querystring": _QUERY_STRING_SERIALIZER,
    "toml": _TOML_SERIALIZER,
    "yaml": _YAML_SERIALIZER,
    "yml": _YAML_SERIALIZER,
    "xml": _XML_SERIALIZER,
}

_SERIALIZERS_EXTENSIONS = [f".{extension}" for extension in _SERIALIZERS.keys()]


def get_format_by_path(path):
    path = path.lower()
    for extension in _SERIALIZERS_EXTENSIONS:
        if path.endswith(extension):
            return extension[1:]
    return None


def get_serializer_by_format(format):
    format_key = (format or "").lower().strip()
    format_key = re.sub(r"[\s\-\_]*", "", format_key)
    return _SERIALIZERS.get(format_key)


def get_serializers_extensions():
    return list(_SERIALIZERS_EXTENSIONS)


def autodetect_format(s):
    if isinstance(s, str) and (is_url(s) or validate_file(s)):
        return get_format_by_path(s)
    return None


def decode(
        s: str,
        format: str,
        **kwargs: Any
    ):
    serializer = get_serializer_by_format(format)
    if not serializer:
        raise ValueError(f"Invalid format: {format}.")
    decode_opts = kwargs.copy()
    data = s.strip() if isinstance(s, str) else s
    data = serializer.decode(s, **decode_opts)
    return data


def encode(
        d: dict,
        format: str,
        **kwargs: Any
    ):
    serializer = get_serializer_by_format(format)
    if not serializer:
        raise ValueError(f"Invalid format: {format}.")
    s = serializer.encode(d, **kwargs)
    return s


def is_data(s):
    if isinstance(s, dict):
        return True
    else:
        return len(s.splitlines()) > 1


def validate_file(s, thrown_error: bool=False):
    filepath = Path(s)
    if filepath.exists():
        if filepath.is_file():
            file = str(s)
            if any([file.endswith(ext)
                    for ext in get_serializers_extensions()]):
                return True
            elif thrown_error:
                raise RuntimeError(f'Unsupported file extension: {filepath}')
        else:
            raise RuntimeError(f'filepath is not file: {filepath}')
    elif thrown_error:
        raise FileNotFoundError(f'No such file or directory: {filepath}')

    return False


def is_url(s):
    return any([s.startswith(protocol) for protocol in ["http://", "https://"]])


def read_contents(s, thrown_error: bool=False):
    # s -> filepath or url or data
    if is_data(s): # data
        return s
    elif is_url(s): # url
        return read_url(s)
    elif validate_file(s, thrown_error=thrown_error):
        return read_file(s, thrown_error=thrown_error)
    # one-line data?!
    return s


def read_url(
        url: str,
        **options: Any
    ):
    if not requests_installed:
        raise ModuleNotInstalledError("'requests' module is not installed.")

    response = requests.get(url, **options)
    response.raise_for_status()
    content = response.text
    return content


def read_file(
        filepath: Union[str, Path],
        encording: str="utf-8",
        thrown_error: bool=False,
        **options: Any
    ):
    contents = ""
    if validate_file(filepath, thrown_error=thrown_error):
        ops = dict(encording=encording)
        ops.update(options)
        with open(filepath, 'r', **options) as file:
            contents = file.read()
    return contents


def write_file(
        filepath: Union[str, Path],
        content: str,
        append: bool=False,
        encording: str="utf-8",
        **options: Any
    ):
    filepath = Path(filepath).absolute()
    this_dir = filepath.parent
    filepath.mdir(parents=True, exists_ok=True)
    mode = 'a' if append else 'w'
    ops = dict(encording=encording)
    ops.update(options)
    with open(str(filepath), mode, **options) as file:
        file.write(content)



# -*- coding: utf-8 -*-

from __future__ import annotations
import copy

from typing import (
    Any, Dict, Union, Optional, Hashable, Iterable, Type, Sequence,
    Callable, Pattern, Match, Literal, get_args
    )

from collections import defaultdict

from datajuggler.utils import StrCase
from datajuggler.keys import (
     Keylist, Keypath, Default_Keypath_Separator, get_index_from_key
     )
from datajuggler.strings import is_match_string, searchstr
from datajuggler.validator import (
    DictAction, DictActionType, validate_DictAction,
    DictItem, DictItemType, validate_DictItem,
    DictKey, DictKeyType, validate_DictKey,
    )
from datajuggler.validator import TypeValidator as _type

def _dict_updator(
        d: dict,
        new: dict,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) -> Optional[dict]:

    if inplace:
        if id(d) != id(new):
            d.clear()
            d.update(new)
    else:
        if not factory or factory == dict:
            return new
        else:
            return factory(new)

def d_clean(
        obj: dict,
        strings=True,
        collections=True,
        *,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ):
    """Clean the current dict instance removing all empty values:
        None, '', {}, [], ().
    If strings or collections (dict, list, set, tuple) flags are False,
    related empty values will not be deleted.
    """

    def _is_empty(obj, key, strings, collections):
        value = obj.get(key, None)
        if not value:
            del_none = value is None
            del_string = strings and _type.is_str(value)
            del_collection = collections and _type.is_collection(value)
            return any([del_none, del_string, del_collection])

        return False

    new = obj if inplace else factory(obj)

    for key in list(obj.keys()):
        if _is_empty(new, key, strings, collections):
            del new[key]

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_clone(
        obj: dict,
        empty: bool=False,
        memo: Optional[dict]=None,
    ):
    """Return a clone (deepcopy) of the dict. """
    d = copy.deepcopy(obj, memo)
    if empty:
        d.clear()
    return d

def d_compare(
        d1: dict,
        d2: dict,
        *,
        keys: Optional[Union[Hashable, list, Keylist, Keypath]]=None,
        thrown_error: bool=False,
    ):
    """Compare tow dictionary with keys.
    and return  `True` when equal found values.
    otherwise return `False`.
    if not set keys, just compare two dictionaries,
    """

    def compare_values(d1: dict, d2: dict, keys):
        if _type.is_mapping(d1):
            d1_values = get_values(d1, keys)
            if _type.is_dict(d1_values):
                d1_values = list(d1_values.values())
        else:
            d1_values = d1

        if _type.is_mapping(d2):
            d2_values = get_values(d2, keys)
            if _type.is_dict(d2_values):
                d2_values = list(d2_values.values())
        else:
            d2_values = d2
        return d1_values == d2_values

    if not keys:
        keys = get_keys(d1)
        check_key = d1 == d2
    else:
        if _type.is_keylist_or_keypath(keys):
            d1 = get_values(d1, keys)
            d2 = get_values(d2, keys)
            keys = get_keys(d1)
        else:
            keys = [keys] if _type.is_str(keys) else keys

        check_key = True

    check_value = compare_values(d1, d2, keys)

    if check_key and check_value:
        return True
    else:
        if thrown_error:
            raise ValueError(f'{str(d1)} is not equal {str(d2)}.')
        return False

def d_counts(
        obj: dict,
        pattern: Union[Hashable, Sequence, Pattern],
        *,
        count_for: DictItemType=DictItem.KEY,
        wild: bool=False,
        verbatim: bool=False,
    ) ->Union[int, dict]:
    """Counts of keys or values
       if pass `wild=True`, match substr and ignore_case.
       if pass `verbatim=True`, counts as it is.
    """

    def _counts_of_keys(
            obj: dict,
            keys: Union[Hashable, Sequence],
            wild: bool=False,
            verbatim: bool=False,
        ) ->Union[int, dict]:
        """Counts of keys"""

        if _type.is_str(keys):
            keys = [keys]
            as_int=True
        else:
            as_int=False
        keys = list(keys) if not _type.is_list(keys) else keys

        counts = defaultdict(int)
        all_keys = get_keys(obj)
        for key in keys:
            if wild:
                for k in all_keys:
                    if is_match_string(key, k, wild):
                        if verbatim:
                            counts[k] += 1
                        else:
                            counts[key] += 1
            else:
                counts[key] += all_keys.count(key)

        if as_int:
            return counts[keys[0]]
        else:
            return counts

    def _counts_of_values(
            obj: dict,
            value: Any,
            wild: bool=False,
            verbatim: bool=False,
        ) ->Union[int, dict]:
        """Counts of values"""

        def deep_lookup(
                value: Any,
                obj: Union[dict, list, tuple],
                wild: bool=False,
                verbatim: bool=False,
            ):

            if _type.is_mapping(obj):
                for k, v in obj.items():
                    if _type.is_str(v):
                        if is_match_string(value, v, wild):
                            if verbatim:
                                yield (1, v)
                            else:
                                yield (1, value)

                    elif _type.is_mapping(v):
                        for c, v in deep_lookup(value, v,
                                          wild=wild, verbatim=verbatim):
                            yield (c, v)
                    elif _type.is_list_or_tuple(v):
                        for element in v:
                            for c, v in deep_lookup(value, element,
                                             wild=wild, verbatim=verbatim):
                                yield (c, v)
                    elif value == v:
                        yield (1, v)

            elif _type.is_list_or_tuple(obj):
                for element in obj:
                    for c, v in deep_lookup(value, element,
                                     wild=wild, verbatim=verbatim):
                        yield (c, v)

        counts = defaultdict(int)

        for c, v in deep_lookup(value, obj, wild, verbatim=verbatim):
            counts[v] += c

        return dict(counts)


    validate_DictItem(count_for, thrown_error=True)

    if count_for == DictItem.KEY:
        counts = _counts_of_keys(obj, pattern, wild=wild, verbatim=verbatim)
    else:
        counts = _counts_of_values(obj, pattern, wild=wild, verbatim=verbatim)
    return counts

def d_filter(
        predicate: Callable,
        obj: dict,
        *,
        factory: Type[dict]=dict,
    ):
    """ Create a new dictionary with filter items in dictionary by item.
    Predicate function receives key, value arguments
    and should return a bool value.
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    new = dict()

    for key, val in obj.items():
        if predicate(key, val):
            new[key] = val

    return _dict_updator(obj, new, factory=factory)


def d_groupby(
        seq: list,
        key: Hashable,
        *,
        factory: Type[dict]=dict,
    ) -> dict:
    """
    A groupby operation involves some combination of splitting the object, applying a function, and combining the results. This can be used to group large amounts of data and compute operations on these groups.
    """
    if not _type.is_list(seq):
        raise ValueError("seq should be a list of dicts.")

    grouped = defaultdict(list)
    for element in seq:
        if not _type.is_mapping(element):
            raise ValueError("element should be a dict.")
        group = element.get(key)
        grouped[group].append(element.copy())

    return factory(grouped)


def d_invert(
        obj: dict,
        flat: bool=False,
        *,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) ->dict:
    """
    Return an inverted dict where values become keys and keys become values.
    Since multiple keys could have the same value, each value will be a list of keys.
    If pass `flat=True` each value will be a single value.
    (use this only if values are unique).
    """

    def _invert_item(
            obj: dict,
            key: Hashable,
            value: Any,
            flat: bool=False,
        ):
        if flat:
            obj.setdefault(value, key)
        else:
            obj.setdefault(value, []).append(key)


    def _invert_list(d, key, value, flat):
        for element in value:
            _invert_item(d, key, element, flat)

    new = factory()
    for key, value in obj.items():
        if _type.is_list_or_tuple(value):
            _invert_list(new, key, value, flat)
        else:
            _invert_item(new, key, value, flat)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)


def d_map(
        func: Callable,
        obj: dict,
        *,
        map_for: Optional[DictItemType]=None,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) ->dict:
    """
    Create a new dictionary with apply function to keys/value of dictionary.
    if pass `map_for=None`  apply function to key and value. (default)
    if pass `map_for="key"`  apply function to key.
    if pass `map_for="value"`  apply function to value.
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    new = obj if inplace else factory(obj)

    if not map_for:
        new = dict(map(func, new.items()))
    else:
        validate_DictItem(map_for, thrown_error=True)
        if map_for == DictItem.KEY:
            new = dict(zip(map(func, new.keys()), new.values()))
        else:
            new = dict(zip(new.keys(), map(func, new.values())))

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_merge(
        obj: dict,
        others: Union[dict, list, tuple],
        *,
        overwrite: bool=True,
        concat: bool=False,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) ->dict:
    """
    Merge one or more dictionary objects into obj.
    Sub-dictionaries keys will be merged toghether.
    If pass `overwrite=False`, existing values will not be overwritten.
    If pass `concat=True`, list values will be concatenated toghether.
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    def _merge_dict(
        obj: dict,
        other: dict,
        overwrite: bool=True,
        concat: bool=False
        ):
        for key, value in other.items():
            _merge_item(obj, key, value, overwrite=overwrite, concat=concat)


    def _merge_item(
        obj: dict,
        key: Hashable,
        value: Any,
        overwrite: bool=True,
        concat: bool=False
        ):

        if key in obj:
            item = obj.get(key, None)
            if _type.is_mapping(item) and _type.is_mapping(value):
                _merge_dict(item, value, overwrite=overwrite, concat=concat)
            elif concat and _type.is_list_or_tuple(item):
                item += value
            elif overwrite:
                obj[key] = value
        else:
            obj[key] = value

    new = obj if inplace else factory(obj)

    others = list(others) if _type.is_tuple(others) else others
    others = [others] if not _type.is_list(others) else others
    for other in others:
        _merge_dict(new, other, overwrite=overwrite, concat=concat)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_move(
        obj: dict,
        key_src: Union[str, list, dict],
        key_dest: Optional[Union[str, list]]=None,
        *,
        overwrite: bool=True,
        keep_order: bool=False,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) ->dict:
    """Create new dictionary which Move an item from key_src to key_dst.
    It can be used to rename a key.
    If key_dst exists and pass `overwrite=True`, its value will be overwritten.
    if pass `keep_order=True`, keep ordered of dictionary. (may be slow).
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    def _move(obj, key_src, key_dest, overwrite, keep_order):
        if key_dest == key_src:
            return
        if key_dest in obj and not overwrite:
            raise KeyError(
                f"Invalid key: '{key_dest}', "
                "key already in dict and 'overwrite' is disabled."
            )

        if keep_order:
            keys = [ key_dest if x == key_src else x for x in obj.keys() ]
            vals = list(obj.values())
            obj.clear()
            obj.update(dict(zip(keys, vals)))
        else:
            obj[key_dest] = obj.pop(key_src)

    new = obj if inplace else factory(obj)

    if _type.is_dict(key_src) and key_dest is None:
        for key_src, key_dest in key_src.items():
            _move(new, key_src, key_dest, overwrite, keep_order)
    else:
        _move(new, key_src, key_dest, overwrite, keep_order)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)


def d_rename(
        obj: dict,
        key: Union[Hashable,dict, list],
        key_new: Optional[Hashable]=None,
        case_name: Optional[str]=None,
        *,
        overwrite: bool=False,
        keep_order: bool=False,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) ->dict:
    """Create the new dictionary which is chnaged the key to key_new.
    if key as dictionary {key: key_new}, change key using mapping dictionary.
    If key_dst exists and pass `overwrite=True`, its value will be overwritten.
    if pass `keep_order=True`, keep ordered of dictionary. (may be slow).
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    if case_name:
        c = StrCase()
        if case_name not in c.supported_case:
            supported_case = c.show_supported_case()
            raise ValueError(f'Invalid case_name.\n{supported_case}')

        if ( _type.is_list_or_tuple(key)
            or  _type.is_dict_keys(key) ) and not key_new:
            key = { x: c.convert_case(case_name, x) for x in key }
        elif _type.is_hashable(key) and not key_new:
            key_new = c.convert_case(case_name, key)
    else:
        if _type.is_list(key):
            raise TypeError('Invalid Type. Perhaps you forgot case_name. ')

    return d_move(obj, key, key_new,
                overwrite=overwrite, keep_order=keep_order,
                inplace=inplace, factory=factory)

def d_remove(
        obj: dict,
        keys: Union[list, tuple, Hashable],
        *,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ):
    """Create new dictionary which Remove multiple keys from the dict.
    It is possible to pass a single key or more keys (as list or *args).
    """

    new = obj if inplace else factory(obj)

    keys = [keys] if _type.is_str(keys) else keys
    keys = list(keys) if _type.is_tuple(keys) else keys
    for key in keys:
        new.pop(key)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_nest(
    items: tuple,
    id_key: Union[str, list],
    parent_id_key: Union[str, list],
    children_key: Union[str, list],
    ) -> list:
    """Nest a list of dicts at the given key and return a new nested list
    using the specified keys to establish the correct items hierarchy.
    """

    def _nest_items(nested_items, item, id_key, children_key):
        children_items = nested_items.pop(item[id_key], [])
        item[children_key] = children_items

        for child_item in children_items:
            _nest_items(nested_items, child_item, id_key, children_key)


    if any( [id_key == parent_id_key,
            id_key == children_key,
            parent_id_key == children_key] ):
        raise ValueError("keys should be different.")

    nested_items = d_groupby(items, parent_id_key)
    root_items = nested_items.get(None, [])
    for item in root_items:
        _nest_items(nested_items, item, id_key, children_key)

    return nested_items.get(None)


def d_subset(
        obj: dict,
        keys: Union[str, list, tuple, Hashable],
        *,
        default: Optional[Any]=None,
        use_keypath: bool=False,
        separator: str=Default_Keypath_Separator,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ):
    """
    Return a dict subset for the given keys.
    It is possible to pass a single key or more keys (as list or *args).
    """

    def get_subset(obj, key, val, parents, search_key):
        nonlocal new, use_keypath, separator
        if key == search_key:
            if use_keypath:
                keypath = separator.join(parents)
                new[keypath] = val
            else:
                if not key in new:
                    new[key] = val
                else:
                    raise KeyError(f"Multiple keys founded.'{key}'")

    new = factory()

    keys = [keys] if _type.is_str(keys) else keys
    keys = list(keys) if not _type.is_list(keys) else keys

    for key in keys:
        d_traverse(obj, get_subset, parents=[], search_key=key)

    if len(new) == 0:
        new[key] = default

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_find(
        obj: dict,
        keys: Union[list,Hashable],
        default: Optional[Any]=None,
        first_one: bool=True,
        factory: Type[dict]=dict,
    ) -> Union[Any, dict]:
    """Return the match searching for the given keys.
    if pass `first_one=True`, return first matches.
    If no result found, default value is returned.
    """

    found = defaultdict(list)
    keys = [keys] if _type.is_str(keys) else keys
    keys = list(keys) if not _type.is_list(keys) else keys

    found = factory()
    val = default
    for key in keys:
        if key in obj:
            val =  obj.get(key, default)
            if first_one:
                return val
            else:
                found[key] = val
    return found if not first_one else default

def d_sort(
        obj: dict,
        sort_by: DictItemType=DictItem.KEY,
        reverse: bool=False,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ):
    """Create new dictiionary which is sorted by keys/values.
    `sort_by` accept "key" and "value". default is  "key".
    If pass `reverse=True`, the list will be reversed.
    If pass `inplace=True`, perform operation in-place.
    If set `factory`, create instance of factory class.
    """

    def _items_sorted_by_item_at_index(d, index, reverse):
        return dict(sorted(d.items(), key=lambda item: item[index],
                                      reverse=reverse))

    validate_DictItem(sort_by, thrown_error=True)

    new = obj if inplace else factory(obj)

    if sort_by == DictItem.KEY:
        new = _items_sorted_by_item_at_index(new, 1, reverse)
    else:
        new = _items_sorted_by_item_at_index(new, 0, reverse)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_search(
        obj: dict,
        query: Pattern,
        search_for: DictItemType=DictItem.KEY,
        exact: bool=False,
        separator: str=Default_Keypath_Separator,
        ignore_case: bool=False,
        use_keypath: bool=True,
    ):
    """ Search and return a list of items matching the given query.
    search_for accept 'key', 'value' or others.
    if pass `search_for='all'`, search key and value.
    if pass `ignore_case=True`, match search string with ignore case.
    if pass `use_keypath`, return result with keypath.
    """

    def get_match(
            query: Pattern,
            value: str,
            exact: bool=False,
            ignore_case: bool=False
        ) -> Optional[Match]:
        if exact:
            return query == value
        else:
            if _type.is_str(query) and _type.is_str(value):
                v =  searchstr(query, value, ignore_case=ignore_case)
                return v
            else:
                return None

    def search_item(
            obj: dict,
            key: Hashable,
            value: Any,
            parents: list=[],
            in_keys: bool=False,
            in_values: bool=True,
            exact: bool=False,
            ignore_case: bool=False,
            separator: str=Default_Keypath_Separator,
            use_keypath: bool=True,
        ):

        match_key = in_keys and get_match(query, key, exact, ignore_case)
        match_val = in_values and get_match(query, value, exact, ignore_case)
        if any([match_key, match_val]):
            keypath = Keylist(parents).to_keypath(separator=separator)
            if not use_keypath:
                keypath = str(keypath)
            items[keypath] =  value

    items = defaultdict(str)
    if search_for == DictItem.KEY:
        in_keys = True
        in_values = False
    elif search_for == DictItem.VALUE:
        in_keys = False
        in_values = True
    else:
        in_keys = True
        in_values = True

    d_traverse(obj, search_item,
                  in_keys=in_keys,
                  in_values=in_values,
                  exact=exact,
                  ignore_case=ignore_case,
                  use_keypath=use_keypath)
    return dict(items)

def d_swap(
        obj: dict,
        key1: Hashable,
        key2: Hashable,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) ->Optional[dict]:
    """Swap items values at the given keys."""

    if key1 == key2:
        return obj

    new = obj if inplace else factory(obj)

    val1 = new[key1]
    val1 = val1.copy() if _type.is_mapping(val1) else val1
    val2 = new[key2]
    val2 = val2.copy() if _type.is_mapping(val2) else val2
    new[key1], new[key2] = val2, val1

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_flatten(
        obj: dict,
        separator: str=Default_Keypath_Separator,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) -> dict:
    """Return a new flattened dict using the given separator to join nested dict keys to flatten keypaths."""

    def _flatten_key(
        base_key: Hashable,
        key: Hashable,
        separator: str=Default_Keypath_Separator,
    )-> str:
        if base_key and separator:
            return f"{base_key}{separator}{key}"
        return key

    def _flatten_item(
        obj: dict,
        base_dict: Hashable,
        base_key: Hashable,
        default: Optional[Any]=None,
        separator: str=Default_Keypath_Separator,
    ):
        new_dict = base_dict
        for key in list(obj.keys()):
            new_key = _flatten_key(base_key, key, separator)
            value = obj.get(key, None)
            if _type.is_mapping(value):
                new_value = _flatten_item( value,
                                base_dict=new_dict, base_key=new_key,
                                separator=separator )
                new_dict.update(new_value)
                continue
            if new_key in new_dict:
                raise KeyError(f"Invalid key: '{new_key}', key already in flatten dict.")
            new_dict[new_key] = value
        return new_dict

    new = factory()
    new =  _flatten_item(obj,
                         base_dict=new, base_key="",
                         separator=separator)
    return _dict_updator(obj, new, inplace=inplace, factory=factory)


def d_unflatten(
        obj: dict,
        default: Optional[Any]=None,
        separator: str=Default_Keypath_Separator,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) -> dict:
    """Return a new unflattened dict using the given separator to join nested dict keys to flatten keypaths."""

    def _unflatten_item(key, value, separator):
        keys = key.split(separator)
        if _type.is_mapping(value):
            return (keys, d_unflatten(value, separator=separator))
        return (keys, value)


    new = factory()
    for key in list(obj.keys()):
        value = obj.get(key, default)
        new_keys, new_value = _unflatten_item(key, value, separator)
        new = update_items(new, new_keys, new_value,
                                 inplace=False, factory=factory)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)


def d_traverse(
        obj: Union[dict, list, tuple],
        callback: Callable,
        parents: list=[],
        *args: Any,
        **kwargs: Any,
    ):
    """Traverse dict or list and apply callback function.
    callback function will be called as follows.
      `callback(obj, key, value, parents=parents, *args,  **kwargs)`
      `callback(obj, index, value, parents=parents, *args,  **kwargs)`
    """

    def _traverse_dict(
            obj: dict,
            callback: Callable,
            parents: list=[],
            parents_type: list=[],
            *args: Any,
            **kwargs: Any,
        ):
        for key in list(obj.keys()):
            parents.append(key)
            value = obj.get(key, None)
            callback(obj, key, value, *args, parents=parents, **kwargs)
            d_traverse(value, callback, *args, parents=parents, **kwargs)
            parents.pop()

    def _traverse_list(
            obj: Union[list, tuple],
            callback: Callable,
            parents: list=[],
            parents_type: list=[],
            *args: Any,
            **kwargs: Any,
        ):
        for index, value in list(enumerate(obj)):
            parents.append(f'[{index}]')
            callback(obj, index, value, parents=parents, *args, **kwargs)
            d_traverse(value, callback, *args, parents=parents, **kwargs)
            parents.pop()

    if not _type.is_function(callback):
            raise ValueError("callback argument must be a callable.")
    if _type.is_mapping(obj):
        _traverse_dict(obj, callback, *args, parents=parents, **kwargs)
    elif _type.is_list_or_tuple(obj):
        _traverse_list(obj, callback, *args, parents=parents, **kwargs)


def d_unique(
        obj: dict,
    ) -> list:
    """Return unique values from dict. """

    new = dict(obj)
    values = []

    for key in obj.keys():
        value = new.get(key, None)
        if value in values:
            new.pop(key, None)
            continue
        values.append(value)

    return values

def d_using(
        obj: dict,
        *,
        inplace: bool=False,
        factory: Type[dict]=dict,
        **kwargs: Any,
    ) -> dict:
    """
    Create the new dictionary which expand with extra items defined by kwargs.
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    new = obj if inplace else factory(obj)
    new.update(kwargs)
    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def d_without(
        obj: dict,
        keys: list,
        *,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ) -> dict:
    """
    Create the new dictionary which without the keys named by args.
    If set `factory`, create instance of factory class.
    If set `True` to `inplace`, perform operation in-place.
    """

    new = obj if inplace else factory(obj)
    for key in keys:
        del new[key]
    return _dict_updator(obj, new, inplace=inplace, factory=factory)


def get_parent_item(
        obj: dict,
        keys: list,
    ):
    # TODO:
    pass

def get_item_hierarchy(
        obj: dict,
        keys: list,
    ):
    # TODO:

    def _get_item_keyval(item, key):
        if _type.is_list_or_tuple(item):
            index = key if _type.is_integer(key) else None
            if index is not None:
                return (index, item[index])
            elif _type.is_dict(item):
                return (key, item[key])
            raise KeyError(f"Invalid key: '{key};")

    items = []
    item = obj
    for key in keys:
        try:
            key, val = _get_item_keyval(item, key)
            items.append(item, key, val)
            item = val
        except (IndexError, KeyError):
            items.append((None, None, None))
            break
    return iitems


def get_keys(
        obj: Optional[dict]=None,
        *,
        indexes: bool=False,
        output_as: Optional[DictKey]=None,
        separator: str=Default_Keypath_Separator,
    ) -> list:
    """ Get to get all keys from dictionary as a List
    This method is able to process on nested dictionary.
    if output_as accept "keylist" and "keypath".
    """

    def get_deeply( obj):
        found_keys = list()

        if _type.is_mapping(obj):
            for key, val in obj.items():
                found_keys.append(key)
                found_keys += get_deeply(val)

        elif _type.is_list_or_tuple(obj):
            for element in obj:
                found_keys += get_deeply(element)

        return found_keys


    if not output_as:
        all_keys = get_deeply(obj)
    else:
        validate_DictKey(output_as, thrown_error=True)

        if output_as == DictKey.KEYLIST:
            all_keys = Keylist.keylists(obj, indexes=indexes)
        else:
            all_keys = Keypath.keypaths(obj, indexes=indexes,
                                             separator=separator)

    return all_keys

def get_values(
        obj: Union[dict, Sequence],
        keys: Union[Hashable, Keylist, Keypath],
    ) -> Any:
    """Get the value of key in the objet(s).
    `obj` : dict, dict[dict], dict[list], list[dict]
    return value, list, dict.
    """

    def deep_lookup(
            key: Hashable,
            obj: Union[dict, list, tuple],
        ):
        """Lookup a key in a nested object(s).
        yield a value
        """

        if _type.is_mapping(obj):
            for k, v in obj.items():
                if key == k:
                    yield k, v
                if _type.is_mapping(v):
                    for kk, v in deep_lookup(key, v):
                        yield (kk, v)
                elif _type.is_list_or_tuple(v):
                    for element in v:
                        for kk, v in deep_lookup( key, element):
                            yield (kk, v)

        elif _type.is_list_or_tuple(obj):
            index = get_index_from_key(key)
            if _type.is_not_none(index):
                yield (index, obj[index] )
            else:
                for element in obj:
                    for k, v in deep_lookup(key, element):
                        yield (k, v)

    as_dict = False
    if _type.is_keypath(keys):
        keys = keys.to_keylist().value()
    elif _type.is_keylist(keys):
        keys = keys.value()
    elif _type.is_tuple(keys):
        keys = [keys]
    elif _type.is_str(keys):
        keys = [keys]
    elif not _type.is_list_not_empty(keys):
        keys = list(keys)
        as_dict = True

    values = defaultdict(list)
    v = obj
    k = None
    for key in keys:
        for k, v in deep_lookup(key, v):
            values[k].append(v)

    if as_dict:
        return values
    else:
        return values[k][-1] if k is not None else None

def get_items(
        obj: dict,
        loc: Union[Hashable, Keylist, Keypath],
        value: Any,
        *,
        func: Optional[Callable]=None,
        factory: Type[dict]=dict,
    ):
    """
    Create new dictionary with new key value pair as d[key]=val.
    If set `True` to `inplace`, perform operation in-place.
    otherwise, not modify the initial dictionary.
    obj
        dictionary on which to operate
    loc
        the location of the value. str and keypath, keylist.
        i.e.: { 'a': { 'b1': { 'c1': {'x': 1 },
                               'c2': {'x': 2 }},
                     { 'b2': { 'c1': {'x': 3 },
                               'c2': {'x': 4 }} }}}
        if set Keylist('a', 'b1', 'c1',  'x') to `loc`, val is 1.
        if set Keypath('a.b1.c1.x')  to `loc`, val is 1.
        giving the location of the value to be changed in `obj`.
    value: the value to aplly
    """

    new = factory(obj)

    new = update_items(new, loc, value, func=func,
                        action=DictAction.GET,
                        inplace=False, factory=factory)
    return new

def pop_items(
        obj: dict,
        loc: Union[Hashable, Keylist, Keypath],
        value: Any,
        *,
        func: Optional[Callable]=None,
        factory: Type[dict]=dict,
    ):
    """ Create new dictionary with new key value pair as d[key]=val.
        If set `True` to `inplace`, perform operation in-place.
        otherwise, not modify the initial dictionary.
    obj
        dictionary on which to operate
    loc
        the location of the value. str and keypath, keylist.
        i.e.: { 'a': { 'b1': { 'c1': {'x': 1 },
                               'c2': {'x': 2 }},
                     { 'b2': { 'c1': {'x': 3 },
                               'c2': {'x': 4 }} }}}
        if set Keylist('a', 'b1', 'c1',  'x') to `loc`, val is 1.
        if set Keypath('a.b1.c1.x')  to `loc`, val is 1.
        giving the location of the value to be changed in `obj`.
        if set loc as str and has `.keypath_separator`,
        convert keypath to str using
        `str.split(keys, sep=separator)`.
    value: the value to aplly
    """
    item = update_items(obj, loc, value, func=func,
                        action=DictAction.POP,
                        inplace=True, factory=factory)
    return item

def del_items(
        obj: dict,
        loc: Union[Hashable, Keylist, Keypath],
        *,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ):
    """ Create new dicttionary with the given key(s) removed.
        New dictionary has d[key] deleted for each supplied key.
        If set `True` to `inplace`, perform operation in-place.
        otherwise, not modify the initial dictionary.
    obj
        dictionary on which to operate
    loc
        the location of the value. str and keypath, keylist.
        i.e.: { 'a': { 'b1': { 'c1': {'x': 1 },
                               'c2': {'x': 2 }},
                     { 'b2': { 'c1': {'x': 3 },
                               'c2': {'x': 4 }} }}}
        if set Keylist('a', 'b1', 'c1',  'x') to `loc`, val is 1.
        if set Keypath('a.b1.c1.x')  to `loc`, val is 1.
        giving the location of the value to be changed in `obj`.
    value: the value to aplly
    func:
        the function to apply the object(s)..
    inplace:
        If set `True` to `inplace`, perform operation in-place.
        otherwise, not modify the initial dictionary.
    """

    obj = obj if inplace else factory(obj)

    new = update_items(obj, loc, None, func=None,
                        action=DictAction.DEL,
                        inplace=False, factory=factory)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)

def set_items(
        obj: Union[dict, Sequence],
        loc: Union[Hashable, Keylist, Keypath],
        value: Any,
        func: Optional[Callable]=None,
        factory: Type[dict]=dict,
    ):
    """ Create new dict with new, potentially nested, key value pair
    obj
        dictionary on which to operate
    loc:
        the location of the value. str and keypath, keylist.
        i.e.: { 'a': { 'b1': { 'c1': {'x': 1 },
                               'c2': {'x': 2 }},
                     { 'b2': { 'c1': {'x': 3 },
                               'c2': {'x': 4 }} }}}
        if set Keylist('a', 'b1', 'c1',  'x') to `loc`, val is 1.
        if set Keypath('a.b1.c1.x')  to `loc`, val is 1.
        giving the location of the value to be changed in `obj`.
        if set loc as str and has `.keypath_separator`,
    value: the value to aplly
    func:
        the function to apply the object(s)..
    inplace:
        If set `True` to `inplace`, perform operation in-place.
        otherwise, not modify the initial dictionary.
    """

    new = update_items(obj, loc, value, func=func,
                         factory=factory, inplace=False)
    return _dict_updator(obj, new, inplace=True, factory=factory)


def update_items(
        obj: Union[dict, Sequence],
        loc: Union[Hashable, Keylist, Keypath],
        value: Optional[Any]=None,
        *,
        default: Optional[Any]=None,
        func: Optional[Callable]=None,
        action: DictActionType=DictAction.SET,
        inplace: bool=False,
        factory: Type[dict]=dict,
    ):
    """ Create new dictionary Update value from a nested dictionary.
    Paramters
    ---------
    loc
        the location of the value. str and keypath, keylist.
        i.e.: { 'a': { 'b1': { 'c1': {'x': 1 },
                               'c2': {'x': 2 }},
                     { 'b2': { 'c1': {'x': 3 },
                               'c2': {'x': 4 }} }}}
        if set ['a', 'b1', 'c1',  'x']  as loc,
        giving the location of the value to be changed in d.
        if set loc as str and has `.keypath_separator`,
        convert keypath to str using
        `str.split(keys, sep=separator)`.
    value: the value to aplly
    obj
        dictionary on which to operate
    func
        function to operate on that value

    If loc is not a key in d,
    update_in creates nested dictionaries to the depth
    specified by the keys, with the innermost value set to func(default).

    If set `True` to `inplace`, perform operation in-place.
    otherwise, not modify the initial dictionary.
    """

    validate_DictAction(action, thrown_error=True)

    if _type.is_keylist(loc):
        loc = loc.value()
    elif _type.is_keypath(loc):
        loc = loc.to_keylist().value()
    elif _type.is_str_not_empty(loc):
        loc = Keypath(loc).to_keylist().value()
    if _type.is_tuple(loc):
        loc = list(loc)

    iter_key = iter(loc)
    k = next(iter_key)

    new = inner = factory()
    new.update(obj)

    if _type.is_not_none(value):
        func = func if func else lambda x: value
    else:
        func = func if func else lambda x: x

    for key in iter_key:
        if k in obj:
            obj = obj[k]
            wdict = factory()
            wdict.update(obj)
        else:
            obj = wdict = factory()

        inner[k] = inner = wdict
        k = key

    if k in obj:
        inner[k] = func(obj[k])
        if action == DictAction.GET:
            item = factory(inner)
            return item
        elif action == DictAction.DEL:
            del inner[k]
        elif action == DictAction.POP:
            item = factory(inner)
            del inner[k]
            _dict_updator(obj, new, inplace=True, factory=factory)
            return item
    else:
        if action not in ( DictAction.DEL, DictAction.POP ):
            inner[k] = func(default)

    return _dict_updator(obj, new, inplace=inplace, factory=factory)


def keylists(
        obj: Any,
        indexes: bool=False,
    ) -> list:
    """keylist is the list of key as keys from dict/list."""

    return Keylist.keylists(obj, indexes)


def keypaths(
        obj: dict,
        indexes: bool=False,
        separator: str=Default_Keypath_Separator,
    ) -> str:
    """
    Keypath is the string for  attribute-sytle access to value.
    (dot-notation by default).
    """
    return Keypath.keypaths(obj, indexes, separator=separator)


__all__ = [
    "d_counts",
    "d_compare",
    "d_groupby",
    "d_invert",
    "d_find",
    "d_filter",
    "d_flatten",
    "d_unflatten",
    "d_map",
    "d_merge",
    "d_move",
    "d_remove",
    "d_rename",
    "d_nest",
    "d_search",
    "d_sort",
    "d_subset",
    "d_swap",
    "d_traverse",
    "d_unique",
    "keylists",
    "keypaths",
    "get_keys",
    "get_values",
    "get_items",
    "pop_items",
    "del_items",
    "set_items",
    "update_items",
]




# -*- coding: utf-8 -*-

from __future__ import annotations
from typing import Optional, Union, Any, Hashable
import re

Default_Keypath_Separator='.'

def get_index_from_key(key) -> Optional[int]:
    pattern = re.compile(r'\[(?P<INDEX>-?[0-9]*)\]')
    if isinstance(key, int):
        return key
    match = re.search(pattern, key)
    if match:
        index = match.group('INDEX')
        index = int(index) if index != '' else None
    else:
        index = None

    return index

class Keylist(list):
    def __init__(self,
            keylist: list=[],
            separator: str=Default_Keypath_Separator
        ):
        self.separator = separator
        if keylist:
            keylist = self._normalize(keylist)
        super().__init__(keylist)

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__str__()})'

    def __str__(self):
        return str(list(self))

    def append(self,
            key: Hashable
        ):
        """
        Append key to the end of the keylist.
        """
        self.validate([key])
        super().append(key)

    def find(self,
            val: Union[Any, list, tuple],
        ) -> list:
        """
        Return the list of index that found val in list.
        otherwise return None
        """
        if isinstance(val, (list, tuple)):
            found = [ i for v in val for i, x in enumerate(self) if x == v ]
        else:
            found = [ i for i, x in enumerate(self) if x == val ]

        return found if found else None


    def insert(self,
            index: int,
            key: Hashable
        ):
        """
        Insert object before index.
        """
        self.validate([key])
        super().insert(index, key)

    @staticmethod
    def keylists(
            obj: Any,
            indexes: bool=False,
        ) -> list:
        """keylist is the list of key as keys from dict/list."""

        def _keylist_for_dict(
                obj: dict,
                parent_keys: list,
                indexes: bool=False,
            ) -> list:

            if not isinstance(parent_keys, list):
                parent_keys = list(parent_keys)

            keylist = []
            for key, value in obj.items():
                keys = parent_keys + [key]
                keylist += [keys]
                keylist += _get_keylist(value, keys, indexes)
            return keylist

        def _keylist_for_list(
                obj: list,
                parent_keys: list,
                indexes: bool=False,
            ) ->list:

            if not isinstance(parent_keys, list):
                parent_keys = list(parent_keys)

            keylist = []
            for key, value in enumerate(obj):
                keys = list(parent_keys)
                keys[-1] += f"[{key}]"
                keylist += [keys]
                keylist += _get_keylist(value, keys, indexes)
            return keylist

        def _get_keylist(
                obj: list,
                parent_keys: list,
                indexes: bool=False,
            ) ->list:
            if isinstance(obj, dict):
                return _keylist_for_dict(obj, parent_keys, indexes)
            elif isinstance(obj, list) and indexes:
                return _keylist_for_list(obj, parent_keys, indexes)
            else:
                return []

        return _get_keylist(obj, [], indexes)

    def _normalize(self,
            keylist: list,
            separator: Optional[str]=None,
        )->list:
        separator = separator or self.separator
        trim_index = lambda x: f'[{x}]' if isinstance(x,int) else x
        keylist = [ trim_index(x) for x in keylist ]
        keypath = Keylist.list2path(keylist, separator)
        keylist = keypath.parse_keypath(separator=separator)
        return keylist

    @staticmethod
    def list2path(
            keylists: list,
            separator: str=Default_Keypath_Separator,
        )->Union[list, Keypath]:
        """Convert from keylist to keypath """
        if isinstance(keylists, list) and  not isinstance(keylists[0], list):
           keylists = [keylists]

        trim = lambda x: x.replace(f'{separator}[', '[')
        trim_index = lambda x: f'[{x}]' if isinstance(x,int) else x
        keypaths = [ trim(separator.join(
                          [ trim_index(key) for key in keylist]))
                                      for keylist in keylists  ]
        keypaths = [ Keypath(x, separator) for x in keypaths ]

        if len(keypaths)==1:
            return keypaths[0]
        else:
            return keypaths


    def to_keypath(self,
            keylist: Opyionsl[Union[list, tuple, Keylist]]=None,
            separator: Optional[str]=None,
            as_str: bool=False,
        ) -> Union[Keypath, list]:

        separator = separator or str(self.separator)
        if separator and not isinstance(separator, str):
            raise ValueError('separator must be Hashable or not empty str.')
        keylist = keylist or list(self)
        keypath = Keylist.list2path(keylist, separator)
        if as_str:
            keypath = [ x.value() for x in keypath ]

        if len(keypath) == 1:
            return keypath[0]
        else:
            return keypath

    def update(self,
             keylist: list
        ):
        self.validate(keylist)
        self.__init__(keylist)

    def value(self):
        return list(self)

    def validate(self,
            keylist: Optional[list, tuple, Keylist]=None,
            thrown_error: bool=True,
        ) ->bool:
        keylist = keylist or list(self)
        if not isinstance(keylist, (list, tuple)):
            check =  False
        else:
            check =  all( map(lambda x: x != '' and isinstance(x, Hashable),
                            keylist) )
        if not check and thrown_error:
            raise ValueError('keylist should be list or tuple. '
                             'each element must be Hashable or not empty str.')
        else:
            return check


class Keypath(str):
    keypath: Optional[Union[str, Keypath]]=None

    def __new__(cls,
            keypath: Optional[str]='',
            separator=Default_Keypath_Separator
        ):
        self = super().__new__(cls, keypath)
        if not ( isinstance(separator, str) and separator != ''):
            raise ValueError('separator must be Hashable or not empty str.')
        self.separator = separator
        keypath and self.validate(keypath)
        return self


    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.__str__()}")'

    @staticmethod
    def has_keypath_separator(keypath,
            separator: str=Default_Keypath_Separator,
        ) ->bool:
        if ( keypath and isinstance(separator, str)
             and keypath.find(separator) >= 0 ):
            return True
        else:
            return False

    @staticmethod
    def keypaths(
            obj: dict,
            indexes: bool=False,
            separator: str=Default_Keypath_Separator,
        ) -> Union[Keypath, list]:
        """
        Keypath is the string for  attribute-sytle access to value.
        (dot-notation by default).
        """
        separator = separator or Default_Keypath_Separator
        if not isinstance(separator, str):
            raise ValueError("separator must be a (non-empty) string.")

        key_lists = Keylist.keylists(obj, indexes=indexes)
        key_paths = [ separator.join([f"{key}" for key in key_list])
                                               for key_list in key_lists ]
        key_paths.sort()
        return key_paths

    @staticmethod
    def path2list(
            keypaths: Union[str, list],
            separator: str=Default_Keypath_Separator
        )->list:
        """Convert from keypath to keylist """

        keypaths = [keypaths] if isinstance(keypaths, str) else keypaths
        keylist = [ Keypath.parse_keypath(x,separator=separator)
                    for x in keypaths ]

        if len(keylist)==1:
            return keylist[0]
        else:
            return keylist


    def to_keylist(self,
            keypaths: Optional[Union[str,list,tuple]]=None,
            separator: OPtional[str]=None,
            as_list: bool=False,
        ) ->Union[Keylist, list]:

        separator = separator or self.separator
        keypaths = keypaths or self
        keypaths = [keypaths] if isinstance(keypaths, (str, Keypath)) else keypaths
        keylist = [ self.parse_keypath(x,separator=separator)
                    for x in keypaths ]

        if not as_list:
            keylist = [ Keylist(x,separator) for x in keylist ]

        if len(keylist) == 1:
            return keylist[0]
        else:
            return keylist

    def validate(self,
            keypath:str,
            separator: Optional[str]=None,
            thrown_error: bool=True,
        ) ->bool:
        if keypath and isinstance(keypath, (str, Keypath)):
            return True
        else:
            if thrown_error:
                raise ValueError('keypath should be str and not empty.')
            else:
                return False

    def value(self):
        return str(self)

    def parse_keypath(self,
            keypath: Optional[str]=None,
            separator: Optional[str]=None
        ):
        separator = separator or self.separator
        RE_KEY_INDEX = r"(?:\[[\'\"]*(\-?[\d]+)[\'\"]*\]){1}$"

        def _split_keys(keypath, separator):
            """ Splits keys using the given separator:
            i.e.: 'item.subitem[1]' -> ['item', 'subitem[1]'].
            """
            if separator:
                return keypath.split(separator)
            return [keypath]

        def _split_key_indexes(key):
            """ Splits key indexes:
            i.e.: 'item[0][1]' -> ['item', 0, 1].
            """
            if "[" in key and key.endswith("]"):
                keys = []
                while True:
                    matches = re.findall(RE_KEY_INDEX, key)
                    if matches:
                        key = re.sub(RE_KEY_INDEX, "", key)
                        index = int(matches[0])
                        keys.insert(0, index)
                        continue
                    keys.insert(0, key)
                    break
                try:
                    keys.remove('') # may be int as key of dictionary.
                except ValueError:
                    pass
                return keys
            return [key]

        def _split_keys_and_indexes(keypath, separator):
            """ Splits keys and indexes using the given separator:
            i.e.: 'item[0].subitem[1]' -> ['item', [0], 'subitem', [1]].
            """
            if isinstance(keypath, str):
                keys1 = _split_keys(keypath, separator)
                keys2 = []
                for key in keys1:
                    keys2 += _split_key_indexes(key)
                return keys2
            return [keypath]

        separator = separator or self.separator
        keypath = keypath or Keypath(self, separator)
        if isinstance(keypath, list):
            keys = []
            for key in keypath:
                keys += self.parse_keypath(key, separator)
            return keys
        else:
            return _split_keys_and_indexes(keypath, separator)

__all__ = [
    "Keylist",
    "Keypath",
    "Default_Keypath_Separator",
]



# -*- coding: utf-8 -*-

from datajuggler.core import aDict, uDict, iList
from datajuggler.keys import Keylist, Keypath

from datajuggler.utils import (
    StrCase, urange, omit_values, replace_values,
    split_chunks, add_df, df_compare, rename_duplicates,
)
from datajuggler.strings import (
    searchstr, substr, is_match_string, remove_accents,
)

__all__ = [
    "aDict",
    "uDict",
    "iList",
    "Keylist",
    "Keypath",
    "StrCase",
    "urange",
    "omit_values",
    "replace_values",
    "add_df",
    "df_compare",
    "split_chunks",
    "rename_duplicates",
    "searchstr",
    "substr",
    "is_match_string",
    "remove_accents",
]



# -*- coding: utf-8 -*-

class AbstractSerializer(object):
    """
    This class describes an abstract serializer.
    """

    def __init__(self):
        super().__init__()

    def decode(self, s, **kwargs):
        raise NotImplementedError()

    def encode(self, d, **kwargs):
        raise NotImplementedError()



from enum import Enum
from typing import (
    Any, Dict, Union, Optional, Hashable, Iterable,
    Callable, Pattern, Match, Literal, get_args
)
from collections.abc import (
    Mapping, KeysView, ValuesView, ItemsView, Sequence
)

import re
from datetime import datetime
from decimal import Decimal
from datajuggler.keys import Keylist, Keypath

try:
    from emoji import is_emoji
except ImportError:
    def is_emoji(s: str):
        raise NotImplementedError('You should install emoji.')

class DictKey(str, Enum):
    KEYLIST = "keylist"
    KEYPATH = "keypath"

DictKeyType = Literal[DictKey.KEYLIST, DictKey.KEYPATH]

class DictItem(str, Enum):
    KEY = "key"
    VALUE = "value"

DictItemType = Literal[DictItem.KEY, DictItem.VALUE]

class DictAction(str, Enum):
    SET = "set"
    GET = "get"
    DEL = "del"
    POP = "pop"

DictActionType = Literal[
    DictAction.SET,
    DictAction.GET,
    DictAction.DEL,
    DictAction.POP,
]

def validate_DictKey(
        dictkey: DictKeyType,
        *,
        thrown_error: bool=False,
    ) ->Optional[bool]:
    if dictkey in get_args(DictKeyType):
        return True
    else:
        if not thrown_error:
            return False
        else:
            available_args = [x.value for x in get_args(DictKeyType)]
            raise ValueError(f"DictKey must be '{available_args}'.")

def validate_DictItem(
        dictitem: DictItemType,
        *,
        thrown_error: bool=False,
    ) ->Optional[bool]:
    if dictitem in get_args(DictItemType):
        return True
    else:
        if not thrown_error:
            return False
        else:
            available_args = [x.value for x in get_args(DictItemType)]
            raise ValueError(f"DictItem must be '{available_args}'.")

def validate_DictAction(
        dictaction: DictActionType,
        *,
        thrown_error: bool=False,
    ) ->Optional[bool]:
    if dictaction in get_args(DictActionType):
        return True
    else:
        if not thrown_error:
            return False
        else:
            available_args = [x.value for x in get_args(DictActionType)]
            raise ValueError(f"DictAction must be '{available_args}'.")


class TypeValidator(object):
    regex = re.compile("").__class__
    re_uuid = re.compile(
        "^([0-9a-f]{32}){1}$|^([0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}){1}$",
        flags=re.IGNORECASE,
    )
    re_financial_number = re.compile(r"(?=\d+,\d+).*")


    @classmethod
    def is_bool(cls, obj: Any):
        return isinstance(obj, type(True))

    @classmethod
    def is_collection(cls, obj: Any):
        return isinstance(obj, (dict, set, list, tuple))

    @classmethod
    def is_callable(cls, obj: Any):
        return isinstance(obj, Callable)

    @classmethod
    def is_datetime(cls, obj: Any):
        return isinstance(obj, datetime)

    @classmethod
    def is_decimal(cls, obj: Any):
        return isinstance(obj, Decimal)

    @classmethod
    def is_dict(cls, obj: Any):
        return isinstance(obj, dict)

    @classmethod
    def is_dict_or_other(cls, obj: Any, other: Any):
        return isinstance(obj, (dict, type(other)))

    @classmethod
    def is_dict_and_not_other(cls, obj: Any, other: Any):
        return ( isinstance(obj, dict)
                 and not isinstance(obj, type(other)))

    @classmethod
    def is_dict_keys(cls, obj: Any):
        return isinstance(obj, KeysView)

    @classmethod
    def is_dict_values(cls, obj: Any):
        return isinstance(obj, ValuesView)

    @classmethod
    def is_dict_items(cls, obj: Any):
        return isinstance(obj, ItemsView)

    @classmethod
    def is_dict_or_list(cls, obj: Any):
        return isinstance(obj, (dict, list))

    @classmethod
    def is_dict_or_list_or_tuple(cls, obj: Any):
        return isinstance(obj, (dict, list, tuple))

    @classmethod
    def is_float(cls, obj: Any):
        return isinstance(obj, float)

    @classmethod
    def is_function(cls, obj: Any):
        return callable(obj)

    @classmethod
    def is_hashable(cls, obj: Any):
        return isinstance(obj, Hashable)

    @classmethod
    def is_integer(cls, obj: Any):
        return isinstance(obj, int)

    def is_integer_or_float(cls, obj: Any):
        return isinstance(obj, (int, float))

    @classmethod
    def is_iterable(cls, obj: Any):
        return isinstance(obj, Iterable)

    @classmethod
    def is_json_serializable(cls, obj: Any):
        json_types = (type(None), bool, dict, float, int, list, str, tuple)
        return isinstance(obj, json_types)

    @classmethod
    def is_keylist(cls, obj: Any):
        return isinstance(obj, Keylist)

    @classmethod
    def is_keypath(cls, obj: Any):
        return isinstance(obj, Keypath)

    @classmethod
    def is_keylist_or_keypath(cls, obj: Any):
        return isinstance(obj, (Keylist, Keypath))

    @classmethod
    def is_list(cls, obj: Any):
        return isinstance(obj, list)

    @classmethod
    def is_list_not_empty(cls, obj: Any):
        return obj and isinstance(obj, list)

    @classmethod
    def is_list_or_tuple(cls, obj: Any):
        return isinstance(obj, (list, tuple))

    @classmethod
    def is_list_of_keylists(cls, obj: Any):
        return ( isinstance(obj, list)
                 and len(obj) >= 1
                 and all(map(lambda x: isinstance(x, Keylist), obj)) )

    @classmethod
    def is_list_of_keypaths(cls, obj: Any):
        return ( isinstance(obj, list)
                 and len(obj) >= 1
                 and all(map(lambda x: isinstance(x, Keypath), obj)) )

    @classmethod
    def is_mapping(cls, obj: Any):
        return isinstance(obj, Mapping)

    @classmethod
    def is_match(cls, obj: Any):
        return isinstance(obj, Match)

    @classmethod
    def is_none(cls, obj: Any):
        return obj is None

    @classmethod
    def is_not_none(cls, obj: Any):
        return obj is not None

    @classmethod
    def is_pattern(cls, obj: Any):
        return isinstance(obj, Pattern)

    @classmethod
    def is_regex(cls, obj: Any):
        return isinstance(obj, cls.regex)

    @classmethod
    def is_same_as(cls, obj: Any, other: Any):
        return isinstance(obj, type(other))

    @classmethod
    def is_sequence(cls, obj: Any):
        return isinstance(obj, Sequence)

    @classmethod
    def is_set(cls, obj: Any):
        return isinstance(obj, set)

    @classmethod
    def is_set_not_empty(cls, obj: Any):
        return obj and isinstance(obj, set)

    @classmethod
    def is_slice(cls, obj: Any):
        return isinstance(obj, slice)

    @classmethod
    def is_str(cls, obj: Any):
        return isinstance(obj, str)

    @classmethod
    def is_str_not_empty(cls, obj: Any):
        return obj and isinstance(obj, str)

    @classmethod
    def is_tuple(cls, obj: Any):
        return isinstance(obj, tuple)

    @classmethod
    def is_tuple_not_empty(cls, obj: Any):
        return obj and isinstance(obj, tuple)

    @classmethod
    def is_uuid(cls, obj: Any):
        return obj and isinstance(obj, str) and cls.re_uuid.match(obj)

    @classmethod
    def is_str_alnum(cls, obj: Any):
        def is_alnum(obj):
            try:
                return isinstance(obj, str) and obj.encode('ascii').isalnum()
            except:
                False
        return True if is_alnum(obj) else False

    @classmethod
    def is_str_alpha(cls, obj: Any):
        def is_alpha(obj):
            try:
                return isinstance(obj, str) and obj.encode('ascii').isalpha()
            except:
                False
        return True if is_alpha(obj) else False

    @classmethod
    def is_str_financial_number(cls, obj: Any):
        return obj and isinstance(obj, str) and cls.re_financial_number.match(obj)

    @classmethod
    def is_str_emoji(cls, obj: Any):
        return obj and isinstance(obj, str) and is_emoji(obj)




# -*- coding: utf-8 -*-

import re
from typing import Any, Union, Pattern, Match, Callable, Optional


__RE_FLAGS = [ re.UNICODE, ( re.IGNORECASE + re.UNICODE) ]

# See Also: https://stackoverflow.com/questions/30069846/how-to-find-out-chinese-or-japanese-character-in-a-string-in-python

cjk_ranges = [
  {"from": ord(u"\u3300"), "to": ord(u"\u33ff")},         # compatibility ideographs
  {"from": ord(u"\ufe30"), "to": ord(u"\ufe4f")},         # compatibility ideographs
  {"from": ord(u"\uf900"), "to": ord(u"\ufaff")},         # compatibility ideographs
  {"from": ord(u"\U0002F800"), "to": ord(u"\U0002fa1f")}, # compatibility ideographs
  {'from': ord(u'\u3040'), 'to': ord(u'\u309f')},         # Japanese Hiragana
  {"from": ord(u"\u30a0"), "to": ord(u"\u30ff")},         # Japanese Katakana
  {"from": ord(u"\u2e80"), "to": ord(u"\u2eff")},         # cjk radicals supplement
  {"from": ord(u"\u4e00"), "to": ord(u"\u9fff")},
  {"from": ord(u"\u3400"), "to": ord(u"\u4dbf")},
  {"from": ord(u"\U00020000"), "to": ord(u"\U0002a6df")},
  {"from": ord(u"\U0002a700"), "to": ord(u"\U0002b73f")},
  {"from": ord(u"\U0002b740"), "to": ord(u"\U0002b81f")},
  {"from": ord(u"\U0002b820"), "to": ord(u"\U0002ceaf")}  # included as of Unicode 8.0
]

def is_cjk(char):
  return any([ range["from"] <= ord(char) <= range["to"]
               for range in cjk_ranges ])


def cjk_substrings(string):
      i = 0
      while i<len(string):
        if is_cjk(string[i]):
          start = i
          while is_cjk(string[i]): i += 1
          yield string[start:i]
        i += 1


def searchstr(
        pattern: Pattern,
        string: str,
        wild: bool=False,
        ignore_case: bool=False,
    ) -> Match:
    """ almost same as re.search.
    Scan through string looking for a match to the pattern, returning
    a Match object, or None if no match was found.
    """
    if wild:
        return re.search(pattern, string, flags = __RE_FLAGS[ignore_case])
    else:
        return re.match(pattern, string, flags = __RE_FLAGS[ignore_case])

def substr(
        pattern: Pattern,
        repl: Union[str, Callable],
        string: str,
        count=0,
        ignore_case: bool=False,
    )-> str:
    """ almost same as re.sub.
    Return the string obtained by replacing the leftmost
    non-overlapping occurrences of the pattern in string by the
    replacement `repl`.  `repl` can be either a string or a callable;
    if a string, backslash escapes in it are processed.
    If it is a callable, it's passed the Match object and must return
    a replacement string to be used.
    """
    return re.sub(pattern, repl, string,
                  count=count, flags=__RE_FLAGS[ignore_case])

def is_match_string(
        pattern: Union[str, Pattern],
        string: str,
        wild: bool=False,
    ) -> bool:


    if pattern == string:
        return True
    else:
        if not ( isinstance(string, str)
                 and isinstance(pattern, (str, Pattern))):
            return False

        match = re.search(pattern, string, flags = __RE_FLAGS[wild])
        if match:
            return True
        else:
            return False


def remove_accents(data: Any) -> Any:
    """Return the normal form for a Unicode string
       using canonical decomposition."""

    if isinstance(data, str):
        data = ( normalize("NFD", data)
                 .encode("ascii", "ignore")
                 .decode("ascii") )
    return data

def remove_character(
        string: str,
        punctuations_list: Optional[list]=None
    ) ->str:
    if not stop_words_list:
        return string
    else:
        return "".join([letter for letter in string.lower() if letter not in punctuations_list])

def remove_words(
        string: str,
        stop_words_list: Optional[list]=None
    ) ->str:
    if not stop_words_list:
        return string
    else:
        return " ".join([word for word in give_string.split(" ") if word not in stopwords_list])




# -*- coding: utf-8 -*-

from datajuggler.serializer.abstract import AbstractSerializer

try:
    import plistlib

    class PListSerializer(AbstractSerializer):
        """
        This class describes a p list serializer.
        https://docs.python.org/3/library/plistlib.html
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            kwargs.setdefault("fmt", plistlib.FMT_XML)
            encoding = kwargs.pop("encoding", "utf-8")
            return plistlib.loads(s.encode(encoding), **kwargs)

        def encode(self, d, **kwargs):
            encoding = kwargs.pop("encoding", "utf-8")
            return plistlib.dumps(d, **kwargs).decode(encoding)

except ImportError:

    class PListSerializer(AbstractSerializer):
        """
        This class describes a p list serializer.
        https://docs.python.org/3/library/plistlib.html
        """

        def __init__(self):
            super().__init__()

        def decode(self, s, **kwargs):
            raise NotImplementedError('You should install plistlib.')

        def encode(self, d, **kwargs):
            raise NotImplementedError('You should install plistlib.')
